/**
 * Created by janis dähne.
 */


import {RequestApiPrefix} from "./constants";
import {ServerResponseWithData} from "../types/response";
import {ResponseCodes} from "../helpers/responseCodes";
import logger, {SeverityLevel} from "../helpers/logger";
import axios, {AxiosRequestConfig, AxiosResponse, CancelToken} from 'axios'
import {testingLayerCallerName} from '../constants'


/**
 * a generic get function
 * @param callerName the name of the caller
 * @param apiPathWithoutPrefix the api path the api path to call ... ApiPrefix will be prefixed
 * @param onProgress not working for some requests (all json?) because total is 0 --> we don't know the real size... backend must set this??
 * @returns {Promise<Tout>} rejects the plain axios response on error
 * @private
 */
export async function get<Tout>(callerName: string, apiPathWithoutPrefix: string,
                                onProgress?: (percentage: number) => void
): Promise<Tout> {

  const funcName = '[get]'
  const callerNameWrapped = `[${callerName}]`

  // tslint:disable-next-line
  let config: AxiosRequestConfig = {}

  if (onProgress) {
    config.onDownloadProgress = (progressEvent) => {
      //make sure we get to the 100% (if we get to 100% then we disable cancel)
      onProgress(Math.ceil((progressEvent.loaded * 100) / progressEvent.total))
    }
  }

  return new Promise<Tout>((resolve, reject) => {

    const request = onProgress
                    ? axios.get(RequestApiPrefix + apiPathWithoutPrefix, config)
                    : axios.get(RequestApiPrefix + apiPathWithoutPrefix)

    request.then(res => {

      //xhr gives 2xx status code
      //normally the server answers with 2xx even if the request gives a bad result...
      //this is to known when some connection problems occur

      const data = res.data as ServerResponseWithData<Tout>

      if (!data) {
        reject(res)
        return
      }

      if (data.code !== ResponseCodes.ok && data.code !== ResponseCodes.partialOk) {
        logger.debug(callerName,
                     `${funcName} error on response: ${JSON.stringify(data)}, path: ${apiPathWithoutPrefix}`,
                     SeverityLevel.error
        )
        reject(res)
        return
      }

      logger.debug(callerName, `${funcName} finished`, SeverityLevel.info)

      resolve(data.data)

    })
      .catch(reason => {
        logger.debug(callerName, `${funcName} error: ${reason.message}, path: ${apiPathWithoutPrefix}`,
                     SeverityLevel.error
        )
        reject(reason)
      })
  })
}

export async function create<Tin, Tout>(callerName: string, apiPathWithoutPrefix: string,
                                        postData: Tin
): Promise<Tout> {

  const funcName = '[create]'
  const callerNameWrapped = `[${callerName}]`

  return new Promise<Tout>((resolve, reject) => {
    axios.post(RequestApiPrefix + apiPathWithoutPrefix, postData)
      .then(res => {

        //xhr gives 2xx status code
        //normally the server answers with 2xx even if the request gives a bad result...
        //this is to known when some connection problems occur

        const data = res.data as ServerResponseWithData<Tout>

        if (data.code !== ResponseCodes.ok && data.code !== ResponseCodes.partialOk) {
          logger.debug(callerName,
                       `${funcName} error on response: ${JSON.stringify(data)}, path: ${apiPathWithoutPrefix}`,
                       SeverityLevel.error
          )
          reject(res)
          return
        }

        logger.debug(callerName, `${funcName} finished`, SeverityLevel.info)

        resolve(data.data)

      })
      .catch(reason => {
        logger.debug(callerName, `${funcName} error: ${reason.message}, path: ${apiPathWithoutPrefix}`,
                     SeverityLevel.error
        )
        reject(reason)
      })
  })
}

/**
 * a generic post function
 * @param callerName the name of the caller
 * @param apiPathWithoutPrefix the api path the api path to call ... ApiPrefix will be prefixed
 * @param {Tin} postData the data to post
 * @returns {Promise<Tout>} rejects the plain axios response on error
 */
export async function update<Tin, Tout>(callerName: string, apiPathWithoutPrefix: string,
                                        postData: Tin
): Promise<Tout> {

  const funcName = '[post]'
  const callerNameWrapped = `[${callerName}]`

  return new Promise<Tout>((resolve, reject) => {
    axios.post(RequestApiPrefix + apiPathWithoutPrefix, postData)
      .then(res => {

        //xhr gives 2xx status code
        //normally the server answers with 2xx even if the request gives a bad result...
        //this is to known when some connection problems occur

        const data = res.data as ServerResponseWithData<Tout>

        if (data.code !== ResponseCodes.ok && data.code !== ResponseCodes.partialOk) {
          logger.debug(callerName,
                       `${funcName} error on response : ${JSON.stringify(data)}, path: ${apiPathWithoutPrefix}`,
                       SeverityLevel.error
          )
          reject(res)
          return
        }

        logger.debug(callerName, `${funcName} finished`, SeverityLevel.info)

        if ((data.code === ResponseCodes.ok || data.code === ResponseCodes.partialOk) && callerName === testingLayerCallerName) {

          //for exam mode this is bad because this leaks the release code...
          //so log without displaying in the ui
          logger.debugSilent(callerName, `${funcName} --TESTING-- REQUEST: ${JSON.stringify(
            postData)} --- RESPONSE: ${JSON.stringify(data)}`, SeverityLevel.debug)
        }

        resolve(data.data)

      })
      .catch(reason => {
        logger.debug(callerName, `${funcName} error: ${reason.message}, path: ${apiPathWithoutPrefix}`,
                     SeverityLevel.error
        )
        reject(reason)
      })
  })
}

export async function rawResponsePost<Tin>(callerName: string, apiPathWithoutPrefix: string,
                                           postData: Tin
): Promise<AxiosResponse> {
  const funcName = '[post]'
  const callerNameWrapped = `[${callerName}]`

  return new Promise<any>((resolve, reject) => {

    axios.post(RequestApiPrefix + apiPathWithoutPrefix, postData)
      .then(res => {
        resolve(res)
      })
      .catch(reason => {
        logger.debug(callerName, `${funcName} error: ${reason.message}, path: ${apiPathWithoutPrefix}`,
                     SeverityLevel.error
        )
        reject(reason)
      })


  })
}


/**
 * a generic delete function (delete is an identifier)
 * deletes only works via url, no post data
 * @param {string} callerName
 * @param {string} apiPathWithoutPrefix
 * @returns {Promise<Tout>}
 */
export async function del<Tout>(callerName: string, apiPathWithoutPrefix: string): Promise<Tout> {

  const funcName = '[delete]'
  const callerNameWrapped = `[${callerName}]`

  return new Promise<Tout>((resolve, reject) => {
    axios.delete(RequestApiPrefix + apiPathWithoutPrefix)
      .then(res => {

        //xhr gives 2xx status code
        //normally the server answers with 2xx even if the request gives a bad result...
        //this is to known when some connection problems occur

        const data = res.data as ServerResponseWithData<Tout>

        if (data.code !== ResponseCodes.ok && data.code !== ResponseCodes.partialOk) {
          logger.debug(callerName,
                       `${funcName} error on response: ${JSON.stringify(data)}, path: ${apiPathWithoutPrefix}`,
                       SeverityLevel.error
          )
          reject(res)
          return
        }

        logger.debug(callerName, `${funcName} finished`, SeverityLevel.info)

        resolve(data.data)

      })
      .catch(reason => {
        logger.debug(callerName, `${funcName} error: ${reason.message}, path: ${apiPathWithoutPrefix}`,
                     SeverityLevel.error
        )
        reject(reason)
      })
  })
}


/**
 *
 * @param callerName
 * @param apiPathWithoutPrefix
 * @param file
 * @param onUploadProgress
 * @param cancelToken create with tokenSource = axios.CancelToken.source() and pass tokenSource.token and then call tokenSource.cancel('upload canceled')
 */
export async function uploadFileAsync<TOut>(callerName: string, apiPathWithoutPrefix: string, file: File,
                                            cancelToken: CancelToken,
                                            onUploadProgress: (percentage: number) => void,

): Promise<TOut> {

  return new Promise<TOut>((resolve, reject) => {

    const formData = new FormData()
    formData.append('file', file)

    const config: AxiosRequestConfig = {
      onUploadProgress: (progressEvent: ProgressEvent) => {
        //make sure we get to the 100% (if we get to 100% then we disable cancel)
        onUploadProgress(Math.ceil((progressEvent.loaded * 100) / progressEvent.total))
      },
      cancelToken: cancelToken
    }

    axios.post(RequestApiPrefix + apiPathWithoutPrefix, formData, config)
      .then((res) => {

        const data = res.data as ServerResponseWithData<TOut>

        if (data.code !== ResponseCodes.ok && data.code !== ResponseCodes.partialOk) {
          logger.debug(callerName, `upload error on response: ${JSON.stringify(data)}, path: ${apiPathWithoutPrefix}`,
                       SeverityLevel.error
          )
          reject(res)
          return
        }

        logger.debug(callerName, `upload finished`, SeverityLevel.info)

        resolve(data.data)

      })
      .catch(reason => {

        if (axios.isCancel(reason)) {
          logger.debug(callerName, `upload canceled: ${reason.message}, path: ${apiPathWithoutPrefix}`, SeverityLevel.error)
          reject(reason)
          return
        }

        logger.debug(callerName, `upload error: ${reason.message}, path: ${apiPathWithoutPrefix}`, SeverityLevel.error)
        reject()
      })
  })


}

