/**
 * Created by janis dähne.
 */

import * as genericLayer from './genericLayer'
import {ExercisePreviewFromBackend} from "../types/exercisePreview";
import {BaseUserData, EmailAndPasswordTuple, IdAndPasswordTuple, SystemRoleFullBase} from "../types/userData";
import {
  NewUserFrontendExclusive, SelectableUserFrontendExclusive, UserPreviewFromBackend,
  UserSystemRoleTupleForBackend
} from "../types/users";
import logger, {SeverityLevel} from '../helpers/logger'
import {PaginatedData, PaginationPostData} from '../types/pagination'

const callingName = 'usersLayer'
const controllerPrefix = 'users'


export function getActivatedUsers(paginationPostData: PaginationPostData, viaGroupId: number | null): Promise<PaginatedData<UserPreviewFromBackend>> {
  return genericLayer.update<PaginationPostData, PaginatedData<UserPreviewFromBackend>>(
    callingName, `${controllerPrefix}/activated/${viaGroupId}`, paginationPostData
  )
}

export function getNotActivatedUsers(paginationPostData: PaginationPostData): Promise<PaginatedData<UserPreviewFromBackend>> {
  return genericLayer.update<PaginationPostData, PaginatedData<UserPreviewFromBackend>>(
    callingName, `${controllerPrefix}/notactivated`, paginationPostData
  )
}

/**
 * updates the given user data
 * @param userData the user data
 * @returns {Promise<BaseUserData>} //backend returns Promise<BaseUserData>
 */
export function adminUpdateUserData(userData: BaseUserData): Promise<void> {
  return genericLayer.update<BaseUserData, void>(callingName,
    `${controllerPrefix}/data/force`, userData)
}

/**
 * resets the password for the given user and returns the generated password
 * @param tuple the target user
 * @returns {Promise<string>}
 */
export function resetPassword(tuple: IdAndPasswordTuple): Promise<string> {

  //not used yet because we always want the server to create a random password
  tuple = {
    ...tuple,
    password: ''
  }

  return genericLayer.update<IdAndPasswordTuple, string>(callingName,
    `${controllerPrefix}/password/reset`, tuple)
}


/**
 * deletes an active user
 * @param user the user
 * @returns {Promise<void>}
 */
export function deleteActive(user: UserPreviewFromBackend): Promise<void> {
  return genericLayer.del(callingName, `${controllerPrefix}/activated/${user.id}`)
}

/**
 * deletes a not active user
 * @param user the user to delete
 * @returns {Promise<void>}
 */
export function deleteNotActive(user: UserPreviewFromBackend): Promise<void> {
  return genericLayer.del(callingName, `${controllerPrefix}/notactivated/${user.id}`)
}

/**
 * activates the given user and sets the given system role
 * @param activateTuple the tuple
 * @returns {Promise<void>}
 */
export function activateUser(activateTuple: UserSystemRoleTupleForBackend): Promise<void> {
  return genericLayer.update(callingName, `${controllerPrefix}/activate`, activateTuple)
}

/**
 * changes the given users system role
 * @param tuple the user and system role
 * @returns {Promise<void>}
 */
export function changeUsersSystemRole(tuple: UserSystemRoleTupleForBackend): Promise<void> {
  return genericLayer.update(callingName, `${controllerPrefix}/activated/systemrole`, tuple)
}

/**
 * the user can only change it's own email and password but not other basic data...
 * @param tuple the data
 * @returns {Promise<BaseUserData>} backend returns BaseUserData but we don't use it
 */
export function updateOwnUserData(tuple: EmailAndPasswordTuple): Promise<BaseUserData> {
  return genericLayer.update<EmailAndPasswordTuple, BaseUserData>(callingName,
    `${controllerPrefix}/data`, tuple)
}

export function toSelectableUser(users: ReadonlyArray<UserPreviewFromBackend>): ReadonlyArray<NewUserFrontendExclusive> {
  return users.map(user => {
    return {
      ...user,
      uiIsSelected: false,
      systemRole: null,
      uiIsChangeSystemRolePopupDisplayed: false
    }
  }) as ReadonlyArray<NewUserFrontendExclusive>
}

/**
 * transforms the given users to selectable users
 * @param users the users
 * @param systemRoles all system roles
 * @param currentUserId the current user id to set isCurrentUser prop
 */
export function toSelectableUsersWithSystemRoles(users: ReadonlyArray<UserPreviewFromBackend>,
                                                 systemRoles: ReadonlyArray<SystemRoleFullBase>,
                                                 currentUserId: number): ReadonlyArray<SelectableUserFrontendExclusive> {

  const usersForFrontend: ReadonlyArray<SelectableUserFrontendExclusive> = users.map((user) => {

    const systemRole = systemRoles.find(p => p.id === user.systemRoleId)

    const selectableUser: SelectableUserFrontendExclusive = {
      ...user,
      isCurrentUser: user.id === currentUserId,
      uiIsSelected: false,
      isDisplayed: true,
      systemRole: (systemRole ? systemRole : null), //find will return undefined we want null
    }

    if (user.systemRoleId > 0 && !systemRole) {
      //user has a system role but could not be found...
      logger.error(`could not find systemRole for user with id ${user.id}` +
        `and systemrole ${user.systemRoleId}`)
    }

    return selectableUser
  })

  return usersForFrontend

}
