/**
 * Created by janis dähne.
 */
import {HasFileChangedMap} from "../../helpers/editorHelper";
import {CSSProperties} from 'react'


/**
 * saves an ace edit session (the undo manager, selected text...)
 * this is needed in case the ace editor is not longer rendered (other component is displayed)
 * and then we need to rerender the whole editor (all session in the editor were removed)
 * so we save the sessions in the tabs
 */
export interface AceEditSession {
  readonly selection: any //too complex
  readonly value: string
  readonly history: {
    readonly undoStack: ReadonlyArray<any>,
    readonly redoStack: ReadonlyArray<any>,
  },
  readonly scrollTop: number,
  readonly scrollLeft: number,

  //todo at least one dirty set via: tab.editSession['option'+'s'] =
  readonly options: {
    readonly firstLineNumber: number
    readonly indentedSoftWrap: boolean
    readonly mode: string
    readonly newLineMode: "auto" | string
    readonly overwrite: boolean
    readonly tabSize: number
    readonly useSoftTabs: boolean
    readonly useWorker: boolean
    readonly wrap: "free" | "off" | string
    readonly wrapMethod: "auto" | string

  }, //really present??
}

export interface EditorTab {
  /**
   * the id of the original file
   */
  readonly id: number
  /**
   * the file name with extension
   */
  readonly fileNameWithExtensions: string
  /**
   * the content
   */
  readonly content: string
  /**
   * is this the main file?
   */
  readonly isMainFile: boolean
  /**
   * true: readonly,
   * false: not
   */
  readonly isReadonly: boolean

  /**
   * true: content is visible for the user,
   * false: the content should be empty and readonly!
   *
   * WE use this prop only to store information, if this is false editor is readonly!
   */
  readonly isContentVisibleForUser: boolean
  /**
   * > 0 file is displayed, < 0 file is not displayed
   *
   * because this is not really used yet 0 is also displayed
   *
   * NOT USED YET
   */
  readonly displayIndex: number
  /**
   * the language
   */
  readonly language: string

  // /**
  //  * the ace edit session so that we reconstruct all sessions & states from a blank view (e.g. because login was displayed)
  //  */
  // readonly editSession?: AceEditSession | null
}


export interface MultiTabCodeEditorProps {

  readonly style?: CSSProperties
  readonly className?: string
  /**
   * a unique id for the editor
   */
  readonly id: string


  //readonly editorToUse: 'monaco' | 'ace'

  readonly height: string

  /**
   * true: at least one files has some unsaved changes
   */
  readonly hasUnsavedChanges: boolean

  /**
   * the parameter is actually every time true.... else this wouldn't be called
   */
  readonly onHasUnsavedChangesChanged: (hasChanges: boolean) => void


  readonly selectedTabId: number
  readonly onSelectedTabIdChanged: (tabId: number) => void

  /**
   * true: the explorer is displayed,
   * false: not
   */
  readonly isFileExplorerDisplayed: boolean
  readonly onChangeIsFileExplorerDisplayed: (isDisplayed: boolean) => void

  /**
   * the file name if the user creates a file
   */
  readonly newFileName: string
  readonly onNewFileNameChanged: (name: string) => void

  /**
   * the id of the file to rename
   *
   * we could use just one prop but maybe we want to change this in future versions...
   */
  readonly renamingFileId: number
  readonly onRenamingFileIdChanged: (fileId: number) => void

  /**
   * true: display file renaming
   */
  readonly isRenamingFile: boolean
  readonly onIsRenamingFileChanged: (isRenaming: boolean) => void
  /**
   * the new file name for binding to the input
   */
  readonly renamingFileName: string

  /**
   * called when renamingFileName should change
   */
  readonly onRenamingFileNameChanged: (newName: string) => void

  /**
   * called when a tab should be opened (e.g. through the explorer)
   */
  readonly onOpenTabClicked: (editorTab: EditorTab) => void

  /**
   * called when a tab is closed through the tabs line
   */
  readonly onCloseTabClicked: (editorTab: EditorTab) => void

  /**
   * can the user create / delete files?
   */
  readonly userCanCreateNewFiles: boolean

  /**
   * called when the user created a new file (tab)
   *
   * yout might want to call setMainFileId (if it's the first file)
   * and openTabChangeXXX file to open the created file (AFTER calling function to add the file)
   *
   * normally mapped to the same function as
   * onRemoveFile, onChange, onCreateFileClicked
   *
   * USE THE HELPER FUNCTION setSolutionFiles_AndRefreshEditor (e.g. with filter) to set the new files
   */
  readonly onCreateFileClicked: (editorTab: EditorTab) => void

  /**
   * called when a file gets renamed (editorTab has the OLD name!)
   *
   * you might want to call setIsRenamingXXXFile and setRenamingXXXFileName to clear the renaming field
   *
   * normally mapped to the same function as
   * onRemoveFile, onChange, onCreateFileClicked
   *
   */
  readonly onFileRename: (editorTab: EditorTab, newName: string) => void

  /**
   * normally mapped to the same function as
   * onRemoveFile, onChange, onCreateFileClicked
   *
   * USE THE HELPER FUNCTION setSolutionFiles_AndRefreshEditor (e.g. with filter) to set the new files
   */
  readonly onRemoveFile: (editorTab: EditorTab) => void

  readonly editorTabs: ReadonlyArray<EditorTab>
  // readonly onChange: (editorTab: EditorTab, newValue: string, changes: any) => void

  /**
   * the value is newer than the editor tabs value
   * the editor tab is the real editor tab, because we use blur the editor tab value is outdated
   */
  readonly onBlur: (value: string, editorTab: EditorTab) => void | null

  /**
   * if set the onBlur will be called debounced with this value
   * cannot be changed after the component did mount!
   */
  readonly debounceTimeInMs?: number

  /**
   * called when the session of the tab is changed (set through editor e.g. initial)
   */
  // readonly onSomeEditorSessionChanged: (changeEditorTabs: ReadonlyArray<EditorTab>) => void

  /**
   * the map to track which file changed (id -> file has changed?)
   */
  readonly hasFileChangedMap: HasFileChangedMap

  /**
   * for this the func editorHelper.updateHasFileChangedMap can be used
   *
   * normally usage:
   *
   * onUpdateHasFileChangedMap={(changedFile: EditorTab, allTabs: ReadonlyArray<EditorTab>) => {
   *        this.props.setHasFileChangedMapExerciseEditor(updateHasFileChangedMap(changedFile, allTabs, this.props.hasFileChangedMap))
   *      }}
   *
   */
  readonly onUpdateHasFileChangedMap: (activeTab: EditorTab, editorTabs: ReadonlyArray<EditorTab>) => void

  /**
   * true: show print button, false: not
   */
  readonly canPrint: boolean
}
