//see https://gist.github.com/jeffijoe/510f6823ef809e3711ed307823b48c0a
//when using react hooks it might be better with https://dev.to/selbekk/persisting-your-react-state-in-9-lines-of-code-9go ??

/*
The MIT License
Copyright (c) Jeff Hansen 2018 to present.
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

//a bit modified...

import * as React from "react";

type Pos = {
  x: number
  y: number
}

/**
 * this is shared as default props across all instances
 */
export const memoryStore = {
  _data: new Map<string, Pos>(),
  get(key: string): Pos | null {
    if (!key) {
      return null
    }
    return this._data.get(key) || null
  },
  set(key: string, data: Pos) {
    if (!key) {
      return
    }
    return this._data.set(key, data)
  }
}

export interface ScrollPosManagerProps {
  /**
   * this must be unique!
   * for our use case we use the panel titles... so make sure they are all unique on the screen!!
   */
  scrollKey: string
  /**
   * do not use this
   */
  scrollStore?: typeof memoryStore

  scrolledChild: ((connectScrollTarget: (node: any) => void) => JSX.Element) | null
}


/**
 * the basic idea is that we create a wrapping component
 * when the child is unmounted also the wrapping component, then we save the scroll pos
 * when we mounted we just restore the scroll pos
 *
 * because we want to set the scroll pos we need a ref to the actual dom node...
 * to get this we need wrap our child in a function where we pass the connect function
 * the child then uses this function and passes it to ref so we get the actual dom node
 *
 * i'm not sure why we need these but lets keep them for now
 * componentWillReceiveProps
 * componentDidUpdate
 *
 * @example
 *
 * ScrollPositionManager scrollKey={this.props.title} scrolledChild={(connectScrollTarget) =>
 *         <div ref={connectScrollTarget} className="tab-content-host-content">
 *           {
 *             this.props.children
 *           }
 *         </div>}/>
 *
 */
export default class ScrollPositionManager extends React.Component<ScrollPosManagerProps, any> {

  _target: HTMLElement

  constructor(props: ScrollPosManagerProps) {
    super(props)
    this.connectScrollTarget = this.connectScrollTarget.bind(this)
  }

  connectScrollTarget(node: HTMLElement) {
    this._target = node
  }

  restoreScrollPosition(pos?: Pos) {
    pos = pos || this.props.scrollStore.get(this.props.scrollKey)
    if (this._target && pos) {

      // console.log(`restore scroll pos for ${this.props.scrollKey}`)
      // requestAnimationFrame(() => {
      //   scroll(this._target, pos.x, pos.y)
      // })
      scrollTo(this._target, pos.x, pos.y)
    }
  }

  saveScrollPosition(key?: string) {
    if (!this._target) return

    // console.log(`save scroll pos for ${this.props.scrollKey}`)

    const pos = getScrollPos(this._target)
    key = key || this.props.scrollKey
    this.props.scrollStore.set(key, pos)
  }

  componentDidMount() {
    this.restoreScrollPosition()
  }

  componentWillReceiveProps(nextProps: ScrollPosManagerProps) {
    if (this.props.scrollKey !== nextProps.scrollKey) {
      this.saveScrollPosition()
    }
  }

  componentDidUpdate(prevProps: ScrollPosManagerProps) {
    if (this.props.scrollKey !== prevProps.scrollKey) {
      this.restoreScrollPosition()
    }
  }

  componentWillUnmount() {
    this.saveScrollPosition()
  }

  render(): JSX.Element {
    const {scrolledChild = null, ...props} = this.props
    return (
      scrolledChild &&
      // children({
      //   ...props,
      //   connectScrollTarget: this.connectScrollTarget
      // })
      scrolledChild(this.connectScrollTarget)
    )
  }
}

//@ts-ignore
ScrollPositionManager.defaultProps = {
  scrollStore: memoryStore
}

function scrollTo(target: HTMLElement, x: number, y: number) {
  if (target instanceof window.Window) {
    target.scrollTo(x, y)
  } else {
    target.scrollLeft = x
    target.scrollTop = y
  }
}

function getScrollPos(target: HTMLElement): Pos {
  if (target instanceof window.Window) {
    return {
      x: target.scrollX,
      y: target.scrollY
    }
  }

  return {
    x: target.scrollLeft,
    y: target.scrollTop
  }
}
