/**
 * Created by janis dähne.
 */

import * as React from "react";
import {connect} from "react-redux";
import {bindActionCreators, Dispatch} from "redux";
import {returntypeof} from 'react-redux-typescript';
import {RootState} from "../../state/reducers";
import {Input, Icon, InputProps} from "semantic-ui-react";
import {ValidationFunc} from "../../validation/validations";
import {ChangeEvent, CSSProperties, KeyboardEventHandler, SyntheticEvent, TextareaHTMLAttributes} from "react";

//const css = require('./styles.styl');

export interface MyProps {
  readonly placeholder?: string

  readonly value?: string
  /**
   * first arg can be null if we have keydown (to insert tab)
   */
  readonly onChange?: (event: ChangeEvent<HTMLTextAreaElement> | null, value: string) => void

  readonly className?: string
  readonly style?: CSSProperties
  /**
   * usage: <MaterialInput disabled>
   * https://stackoverflow.com/questions/37828543/how-to-pass-props-without-value-to-component
   */
  readonly disabled?: boolean

  /**
   * allows scrolling
   */
  readonly readonly?: boolean

  /**
   * only visually hidden (display: hidden)
   */
  readonly isHidden?: boolean

  readonly resizeVertical?: boolean

  /**
   * the validations for the input
   *
   *  usage:
   *
   * <MaterialInput
   *    validations={validationRules.displayName}
   *    validationMessageKey={validationMessageKeys.displayName}
   *  />
   *
   *  validationRules must be specified in the reducer!!
   */
  readonly validations?: ReadonlyArray<ValidationFunc<any>>

  /**
   * the key of the message when the validation fails
   *
   *
   * usage:
   *
   * <MaterialInput
   *    validations={validationRules.displayName}
   *    validationMessageKey={validationMessageKeys.displayName}
   *  />
   *
   * validationMessageKeys must be specified in the reducer!!
   */
  readonly validationMessageKey?: string

  readonly rows?: number

  readonly enableTabInput?: boolean

}

const mapStateToProps = (rootState: RootState, props: MyProps) => {
  return {
    ...props
    //inputProps: {...props},
    // placeholder: props.placeholder,
    // icon: props.icon,
    // iconClickHandler: props.iconClickHandler,
    // value: props.value,
    // disabled: props.disabled,
  }
}

const mapDispatchToProps = (dispatch: Dispatch) => bindActionCreators({
                                                                        //imported reducer funcs here

                                                                      }, dispatch)


const stateProps = returntypeof(mapStateToProps);
const dispatchProps = returntypeof(mapDispatchToProps);
type Props = typeof stateProps & typeof dispatchProps;

/**
 *
 * a material input component
 * not all original (semantic ui input) props are compatible e.g. label needs to be provided
 *
 * usage:
 * <div>
 *   <label>xxx</label>
 *   <MaterialInput />
 * </div>
 *
 * allow to input tab-characters from https://stackoverflow.com/questions/38385936/change-the-cursor-position-in-a-textarea-with-react
 */
class MaterialTextAreaInput extends React.Component<Props, any> {

  inputEl: HTMLTextAreaElement | null = null

  render(): JSX.Element {
    return (
      <div className={`material-textarea ${this.props.className}`}
           style={this.props.isHidden === true
                  ? {
               ...this.props.style,
               visibility: 'hidden'
             }
                  : {...this.props.style}}>
        {
          //note that we pass all props here with
          //https://zhenyong.github.io/react/docs/transferring-props.html or
          //https://zhenyong.github.io/react/docs/jsx-spread.html
          //so we need to reset the className here else it would override the className set in this code
        }
        <textarea
          ref={(e) => this.inputEl = e}
          className={`${this.props.resizeVertical
                        ? 'v-resize'
                        : ''}`}
          readOnly={this.props.readonly}
          placeholder={(this.props.placeholder
                        ? this.props.placeholder
                        : null)}
          onKeyDown={(e: React.KeyboardEvent<HTMLTextAreaElement>) => {
            if (!this.props.enableTabInput) return

            if (e.key.toLowerCase() === 'tab') {
              e.preventDefault()

              // @ts-ignore
              const {selectionStart, selectionEnd} = e.target

              let val = this.props.value === undefined || this.props.value === null
                        ? ''
                        : '' + this.props.value

              const newValue =
                val.substring(0, selectionStart) +
                "\t" +
                val.substring(selectionEnd)

              this.props.onChange(null, newValue)

              if (this.inputEl) {
                this.inputEl.value = newValue
                this.inputEl.selectionStart = this.inputEl.selectionEnd =
                  selectionStart + 1
              }
            }
          }}
          onChange={(e: ChangeEvent<HTMLTextAreaElement>) => {
            this.props.onChange(e, e.target.value)
          }}
          disabled={(this.props.disabled)} value={this.props.value
                                                  ? this.props.value
                                                  : ''}
          rows={this.props.rows}
        />
        <span
          className={
            (this.props.validations && !this.props.validations.every(
              (validation) => validation(this.props.value)))
            ? 'bar error-bar'
            :
            'bar'
          }
        >
          </span>
        {
          this.props.validations &&
          <div className="error-colored brake-words">
            {
              !this.props.validations.every((validation) => validation(this.props.value)) &&
              this.props.validationMessageKey
            }
          </div>
        }

      </div>
    )
  }
}

export default connect(mapStateToProps, mapDispatchToProps)(MaterialTextAreaInput)
