/**
 * Created by janis dähne.
 */

import * as React from "react";
import Logger from "../../../helpers/logger";
import {isReactElement, joinWith, nameOf, getPropertyName} from "../../../helpers/typeHelpers";
import Panel from "../panel/panel";
import {CSSProperties, DragEventHandler} from "react";
import {MouseEvent} from 'react'
import {throttle} from 'lodash-es'

//const css = require('./styles.styl');


/**
 * the different panel view states (sizes)
 */
export enum PanelViewSize {

  normal,

  verticalFullSized,
  horizontalFullSized,

  fullSized,

}

interface MyProps {
  readonly topLeftPanel?: JSX.Element
  readonly topRightPanel?: JSX.Element
  readonly bottomLeftPanel?: JSX.Element
  readonly bottomRightPanel?: JSX.Element

  readonly width: string
  readonly height: string

  /**
   * the horizontal offset 0 --> 50% divider pos left
   *
   * this will be calced in css like left: calc(50% + hDividerPosOffset)
   */
  readonly hDividerPosOffset: number
  /**
   * the horizontal offset 0 --> 50% divider pos left
   *
   * this will be calced in css like left: calc(50% + hDividerPosOffset)
   */
  readonly vDividerPosOffset: number

  readonly onHDividerPosOffsetChanged: (offset: number) => void
  readonly onVDividerPosOffsetChanged: (offset: number) => void

  readonly topRightPanelViewSize: PanelViewSize
  readonly topLeftPanelViewSize: PanelViewSize
  readonly botRightPanelViewSize: PanelViewSize
  readonly botLeftPanelViewSize: PanelViewSize

}

// noinspection TsLint
let dividerStartPoses = {
  x: 0,
  y: 0,
  xOffsetOnMouseDown: 0,
  yOffsetOnMouseDown: 0,
  isVDividerMouseDown: false,
  isHDividerMouseDown: false,
}

export default class PanelGrid extends React.Component<MyProps, any> {


  updateVDraggingHandleFunc: (e: MouseEvent<HTMLDivElement>) => void
  updateHDraggingHandleFunc: (e: MouseEvent<HTMLDivElement>) => void

  componentWillMount(): void {

    //check if we only use proper child elements, this ensures header works
    React.Children.forEach(this.props.children, (child, index) => {
      if (isReactElement(child)) {

        if ((child && child.type === Panel) || child === null) {
          //ok
        } else {
          Logger.error(`a panel can only contain ${nameOf({Panel})} elements`)
        }
      } else {

        if (!child) {
          //react text element... or undefined...
          Logger.error(`not a react element | ${nameOf({PanelGrid})} can only contain ${nameOf({Panel})} elements`)
        }
      }
    })

    this.updateVDraggingHandleFunc = this.getThrottledVDraggingHandler()
    this.updateHDraggingHandleFunc = this.getThrottledHDraggingHandler()
  }

  getThrottledVDraggingHandler(): (e: MouseEvent<HTMLDivElement>) => void {

    //sparse the browser...
    // noinspection TsLint
    return throttle((e: MouseEvent<HTMLDivElement>) => {
      const deltaY = e.clientY - dividerStartPoses.y
      this.props.onVDividerPosOffsetChanged(dividerStartPoses.yOffsetOnMouseDown + deltaY)
    }, 100)

  }

  getThrottledHDraggingHandler(): (e: MouseEvent<HTMLDivElement>) => void {

    //sparse the browser...
    // noinspection TsLint
    return throttle((e: MouseEvent<HTMLDivElement>) => {
      const deltaX = e.clientX - dividerStartPoses.x
      this.props.onHDividerPosOffsetChanged(dividerStartPoses.xOffsetOnMouseDown + deltaX)
    }, 100)
  }

  render(): JSX.Element {

    // const hDividerPos = '70%' // | divider
    // const vDividerPos = '30%' // ---- divider

    // const hDividerPos = `${this.props.hDividerPosOffset}%` // | divider
    // const vDividerPos = `${this.props.vDividerPosOffset}%` // ---- divider

    const normalZIndex = 1
    const overlapZIndex = 2
    const gripperSize = 6

    //normally nested calcs were allowed... but not in the grammar
    //new browser supports this...
    //but for ff41 e.g. this is not supported (some safe browser uses ff41)
    //https://stackoverflow.com/questions/27597034/calc-function-inside-another-calc-in-css
    const hDividerPosWithoutCalc = `(50% + ${this.props.hDividerPosOffset}px - ${gripperSize}px)`
    const hDividerPos = `calc${hDividerPosWithoutCalc}` // | divider

    const vDividerPosWithoutCalc = `(50% + ${this.props.vDividerPosOffset}px - ${gripperSize}px)`
    const vDividerPos = `calc${vDividerPosWithoutCalc}` // ---- divider



    const hDividerStyle: CSSProperties = {
      position: "absolute",
      top: 0,
      left: hDividerPos,
      //right: `calc(100% - ${hDividerPosWithoutCalc})`,
      bottom: 0,
      width: `${gripperSize}px`,
      //backgroundColor: 'red', //just for debugging
      cursor: 'ew-resize'
    }

    const vDividerStyle: CSSProperties = {
      position: "absolute",
      top: vDividerPos,
      left: 0,
      right: 0,
      //bottom: `calc(100% - ${vDividerPosWithoutCalc})`,
      height: `${gripperSize}px`,
      //backgroundColor: 'blue', //just for debugging
      cursor: 'ns-resize',
    }

    // noinspection TsLint
    let topLeftStyle: CSSProperties = {
      position: "absolute",
      top: 0,
      left: 0,
      right: `calc(100% - ${hDividerPosWithoutCalc})`,
      bottom: `calc(100% - ${vDividerPosWithoutCalc})`,
      zIndex: normalZIndex,
    }

    // noinspection TsLint
    let topRightStyle: CSSProperties = {
      position: "absolute",
      top: 0,
      right: 0,
      left: hDividerPos,
      bottom: `calc(100% - ${vDividerPosWithoutCalc})`,
      zIndex: normalZIndex,
    }

    // noinspection TsLint
    let bottomLeftStyle: CSSProperties = {
      position: "absolute",
      top: vDividerPos,
      left: 0,
      right: `calc(100% - ${hDividerPosWithoutCalc})`,
      bottom: 0,
      zIndex: normalZIndex,
    }

    // noinspection TsLint
    let bottomRightStyle: CSSProperties = {
      position: "absolute",
      top: vDividerPos,
      right: 0,
      left: hDividerPos,
      bottom: 0,
      zIndex: normalZIndex,
    }

    //--- full sized
    if (this.props.topLeftPanelViewSize === PanelViewSize.fullSized) {
      topLeftStyle = {
        ...topLeftStyle,
        zIndex: overlapZIndex,
        right: '0',
        bottom: '0'
      }
    }

    if (this.props.topRightPanelViewSize === PanelViewSize.fullSized) {
      topRightStyle = {
        ...topRightStyle,
        zIndex: overlapZIndex,
        left: '0',
        bottom: '0'
      }
    }

    if (this.props.botLeftPanelViewSize === PanelViewSize.fullSized) {
      bottomLeftStyle = {
        ...bottomLeftStyle,
        zIndex: overlapZIndex,
        right: '0',
        top: '0'
      }
    }

    if (this.props.botRightPanelViewSize === PanelViewSize.fullSized) {
      bottomRightStyle = {
        ...bottomRightStyle,
        zIndex: overlapZIndex,
        left: '0',
        top: '0'
      }
    }

    //--- horizontal full sized
    if (this.props.topLeftPanelViewSize === PanelViewSize.horizontalFullSized) {
      topLeftStyle = {
        ...topLeftStyle,
        zIndex: overlapZIndex,
        right: '0',
      }
    }

    if (this.props.topRightPanelViewSize === PanelViewSize.horizontalFullSized) {
      topRightStyle = {
        ...topRightStyle,
        zIndex: overlapZIndex,
        left: '0',
      }
    }

    if (this.props.botLeftPanelViewSize === PanelViewSize.horizontalFullSized) {
      bottomLeftStyle = {
        ...bottomLeftStyle,
        zIndex: overlapZIndex,
        right: '0',
      }
    }

    if (this.props.botRightPanelViewSize === PanelViewSize.horizontalFullSized) {
      bottomRightStyle = {
        ...bottomRightStyle,
        zIndex: overlapZIndex,
        left: '0',
      }
    }

    //---  vertical full sized
    if (this.props.topLeftPanelViewSize === PanelViewSize.verticalFullSized) {
      topLeftStyle = {
        ...topLeftStyle,
        zIndex: overlapZIndex,
        bottom: '0'
      }
    }

    if (this.props.topRightPanelViewSize === PanelViewSize.verticalFullSized) {
      topRightStyle = {
        ...topRightStyle,
        zIndex: overlapZIndex,
        bottom: '0'
      }
    }

    if (this.props.botLeftPanelViewSize === PanelViewSize.verticalFullSized) {
      bottomLeftStyle = {
        ...bottomLeftStyle,
        zIndex: overlapZIndex,
        top: '0'
      }
    }

    if (this.props.botRightPanelViewSize === PanelViewSize.verticalFullSized) {
      bottomRightStyle = {
        ...bottomRightStyle,
        zIndex: overlapZIndex,
        top: '0'
      }
    }

    return (
      <div style={{
        width: this.props.width,
        height: this.props.height
      }}>

        <div className="panel-grid-wrapper">
          <div className="panel-grid"
               onMouseMove={(e: MouseEvent<HTMLDivElement>) => {
                 if (dividerStartPoses.isVDividerMouseDown) {
                   //https://reactjs.org/docs/events.html#event-pooling
                   e.persist()
                   this.updateVDraggingHandleFunc(e)
                 }


                 if (dividerStartPoses.isHDividerMouseDown) {
                   //https://reactjs.org/docs/events.html#event-pooling
                   e.persist()
                   this.updateHDraggingHandleFunc(e)
                 }
               }}
               onMouseUp={() => {
                 dividerStartPoses.isVDividerMouseDown = false
                 dividerStartPoses.isHDividerMouseDown = false
               }}
          >

            <div className="panel-gripper-h-size" style={hDividerStyle}
                 onMouseDown={(e: MouseEvent<HTMLDivElement>) => {
                   dividerStartPoses.x = e.clientX
                   dividerStartPoses.isHDividerMouseDown = true
                   dividerStartPoses.xOffsetOnMouseDown = this.props.hDividerPosOffset
                 }}
                 onMouseUp={() => {
                   dividerStartPoses.isHDividerMouseDown = false
                 }}
            >
            </div>
            <div className="panel-gripper-v-size" style={vDividerStyle}
                 onMouseDown={(e: MouseEvent<HTMLDivElement>) => {
                   dividerStartPoses.y = e.clientY
                   dividerStartPoses.isVDividerMouseDown = true
                   dividerStartPoses.yOffsetOnMouseDown = this.props.vDividerPosOffset
                 }}
                 onMouseUp={() => {
                   dividerStartPoses.isVDividerMouseDown = false
                 }}
            ></div>

            {
              this.props.topLeftPanel &&
              <div style={topLeftStyle} className="panel-wrapper-top-left">
                {
                  this.props.topLeftPanel
                }
              </div>
            }
            {
              this.props.topRightPanel &&
              <div style={topRightStyle} className="panel-wrapper-top-right">
                {
                  this.props.topRightPanel
                }
              </div>
            }
            {
              this.props.bottomLeftPanel &&
              <div style={bottomLeftStyle} className="panel-wrapper-bot-left">
                {
                  this.props.bottomLeftPanel
                }
              </div>
            }
            {
              this.props.bottomRightPanel &&
              <div style={bottomRightStyle} className="panel-wrapper-bot-right">
                {
                  this.props.bottomRightPanel
                }
              </div>
            }


          </div>
        </div>
      </div>
    )
  }
}
