/**
 * Created by janis dähne.
 */

import * as React from "react";
import {connect} from "react-redux";
import {bindActionCreators, Dispatch} from "redux";
import {returntypeof} from 'react-redux-typescript';
import {RootState} from "../../../../state/reducers";

import ListView from './listView'

import SinglePanelSiteWrapper from '../../../singlePanelSiteWrapper'
import MaterialSearchInput from '../../../material/materialSearchInput'

import EnterReleaseCodeDialog from '../dialogs/enterReleaseCodeDialog'

import {
  getOpenExercisesViaCode,
  loadOpenExercisePreviewsViaCodeAsync,
  setActivePaginationPage,
  setOpenExercisesViaCode,
  setPaginationPageSize,
  setSearchText,
  setSortByKey
} from "../../../../state/actions/allAccessibleExercises/openExercisesSite/openExercisesSiteActions";
import Spinner from "../../../helpers/spinner";
import {Moment} from "moment";
import {getServerTime} from "../../../../communicationLayer/timeLayer";
import {updateExercisePreviews} from "../../../../helpers/editorHelper";
import TagFilterPanel from '../../../tagComponents/tagsFilterPanel'
import {getI18n} from "../../../../../i18n/i18nRoot";
import Logger from '../../../../helpers/logger'
import {InputOnChangeEvent} from '../../../../types/reactEvents'
import * as moment from "moment";
import {warningDialog} from '../../../../helpers/dialogHelper'
import {ErrorHelper} from '../../../../helpers/errorHelper'

//const css = require('./styles.styl');

export interface MyProps {
  //readonly test: string
  /**
   * true: load all data,
   * false: only load exercises
   */
  readonly shouldLoadData: boolean

  /**
   * if we display both via code & via visibility we would get 2 dialogs...
   */
  readonly canDisplayEnterReleaseCodeDialog: boolean

  /**
   * true: can display, false not (probably because parent does)
   */
  readonly canDisplayTagFilterPanel: boolean
}

const mapStateToProps = (rootState: RootState, props: MyProps) => {
  return {
    //test0: rootState...
    //test: props.test
    openExercisePreviewsViaCode: rootState.openExercisesState.openExercisePreviewsViaCode,
    isLoading: rootState.openExercisesState.isLoading || rootState.openExercisesState.isLoadingViaCode
      || rootState.manageTagsState.isLoading,
    isEnterReleaseCodeDialogDisplayed: rootState.openExercisesState.isEnterReleaseCodeDialogDisplayed,
    shouldLoadData: props.shouldLoadData,
    sortByKey: rootState.openExercisesState.paginationViaCode.orderByProp,
    sortDirection: rootState.openExercisesState.paginationViaCode.isOrderByDesc,
    pagination: rootState.openExercisesState.paginationViaCode,
    canDisplayEnterReleaseCodeDialog: props.canDisplayEnterReleaseCodeDialog,

    selectedFilterTagsIds: rootState.tagsFilterState.selectedFilterTagsIds,
    selectedNegativeFilterTagsIds: rootState.tagsFilterState.selectedNegativeFilterTagsIds,

    canDisplayTagFilterPanel: props.canDisplayTagFilterPanel,
    isTagsFilterDisplayed: rootState.openExercisesState.isTagsFilterDisplayed,
    tags: rootState.manageTagsState.tags,
    langId: rootState.i18nState.langId
  }
}

const mapDispatchToProps = (dispatch: Dispatch) => bindActionCreators({
                                                                        //imported reducer funcs here
                                                                        loadOpenExercisePreviewsViaCodeAsync,
                                                                        getOpenExercisesViaCode,
                                                                        setSortByKey,
                                                                        setOpenExercisesViaCode,
                                                                        setActivePaginationPage,
                                                                        setPaginationPageSize,
                                                                        setSearchText
                                                                      }, dispatch)


const stateProps = returntypeof(mapStateToProps);
const dispatchProps = returntypeof(mapDispatchToProps);
type Props = typeof stateProps & typeof dispatchProps;



// noinspection TsLint
let updateLeftTimeIntervals: number = -1

class OpenExercisesViaCodeSite extends React.Component<Props, any> {


  async componentWillMount(): Promise<void> {

    try {
      if (this.props.shouldLoadData) {
        await this.props.loadOpenExercisePreviewsViaCodeAsync()
      } else {
        await this.props.getOpenExercisesViaCode()
      }
    } catch(err) {
      ErrorHelper.makeServerErrorDialog(err, this.props.langId, 'Error', 'Could not load site data')
      return //then do not setup timers...
    }

    await this.setupTimeIsUpIntervals()
  }

  componentWillUnmount(): void {
    clearInterval(updateLeftTimeIntervals)
    updateLeftTimeIntervals = -1
  }

  async setupTimeIsUpIntervals(): Promise<void> {

    // noinspection TsLint
    let serverTime: Moment = moment()

    try {
      serverTime = await getServerTime();
    } catch (err) {
      Logger.error('could not get server time')
      warningDialog('', getI18n(this.props.langId, "Could not get the server time, thus the displayed (left) times are probably wrong. The server is probably offline."), this.props.langId)
    }

    const clientServerTimeDiffInS = moment().diff(serverTime, 'seconds')

    clearInterval(updateLeftTimeIntervals)

    updateLeftTimeIntervals = window.setInterval(() => {

      const copy = updateExercisePreviews(this.props.openExercisePreviewsViaCode, clientServerTimeDiffInS, this.props.langId)

      this.props.setOpenExercisesViaCode(copy)

    }, 1000)
  }

  render(): JSX.Element {

    return (

      <div className="multiple-single-panels">

        {
          (this.props.canDisplayTagFilterPanel && !this.props.isEnterReleaseCodeDialogDisplayed && this.props.isTagsFilterDisplayed) &&
          <TagFilterPanel
            isLoading={this.props.isLoading}
            allTags={this.props.tags}
            onTagsUpdated={this.props.getOpenExercisesViaCode}
          />
        }
        <SinglePanelSiteWrapper>

          {
            this.props.isLoading &&
            <Spinner/>
          }

          {
            !this.props.isEnterReleaseCodeDialogDisplayed &&
            <div className="view-padding">
              <div className="view-head-line">
                <div className="view-caption">
                  <h1>
                    {
                      getI18n(this.props.langId, 'Open exercises via code')
                    }
                  </h1>
                </div>
                <div className="view-options">
                  <MaterialSearchInput
                    value={this.props.pagination.searchText}
                    onChange={(e: InputOnChangeEvent) => {
                      this.props.setSearchText(e.currentTarget.value, true)
                    }}
                  />
                </div>
              </div>

              <div className="view-content">
                  <ListView
                    pagination={this.props.pagination}
                    setActivePaginationPage={this.props.setActivePaginationPage}
                    setPaginationPageSize={this.props.setPaginationPageSize}
                    exercisePreviews={this.props.openExercisePreviewsViaCode}
                    displayAsViaCode={true}
                    sortByKey={this.props.sortByKey}
                    sortDirection={this.props.sortDirection}
                    setSortByKey={this.props.setSortByKey}
                    hideShouldNotCountColumn={false}
                  />
              </div>
            </div>
          }

          {
            (!this.props.isLoading && this.props.isEnterReleaseCodeDialogDisplayed && this.props.canDisplayEnterReleaseCodeDialog) &&
            <EnterReleaseCodeDialog/>
          }


        </SinglePanelSiteWrapper>

      </div>

    )
  }
}

export default connect(mapStateToProps, mapDispatchToProps)(OpenExercisesViaCodeSite)
