/**
 * Created by janis dähne.
 */

import * as React from "react";
import {connect} from "react-redux";
import {bindActionCreators, Dispatch} from "redux";
import {returntypeof} from 'react-redux-typescript';
import {RootState} from "../../../../state/reducers";
import Spinner from "../../../helpers/spinner";
import SinglePanelSiteWrapper from '../../../singlePanelSiteWrapper'
import {CartesianGrid, Line, LineChart, ResponsiveContainer, Tooltip, XAxis, YAxis} from "recharts";
import {displayDate} from "../../../../helpers/convertersAndTransformers";
import {SingleAssessmentStatistic} from "../../../../types/assessmentStatistics";
import {AssessmentTooltip} from "./assessmentTooltip";
import {blueColor, errorColor, greenColor} from "../../../../styles/styles";
import {Checkbox, Form} from "semantic-ui-react";
import {
  setDisplayManualPointsLine,
  setDisplayNormalTestLine,
  setDisplaySubmitTestLine
} from "../../../../state/actions/assessmentStatisticsSite/assessmentStatisticsSiteActions";
import {getI18n} from "../../../../../i18n/i18nRoot";
import {HelpPopup} from '../../../helpers/helpPopup'

//const css = require('./styles.styl');

export interface MyProps {
  //readonly test: string
}

const mapStateToProps = (rootState: RootState /*, props: MyProps*/) => {
  return {
    //test0: rootState...
    //test: props.test
    isLoading: rootState.assessmentStatisticsSiteState.isLoading
      || rootState.pLangsState.isLoading,

    assessmentStatistics: rootState.assessmentStatisticsSiteState.assessmentStatistics,
    assessmentStatisticsList: rootState.assessmentStatisticsSiteState.assessmentStatistics
                              ?
                              rootState.assessmentStatisticsSiteState.assessmentStatistics.assessmentStatistics
                              : [],

    displaySubmitTestLine: rootState.assessmentStatisticsSiteState.displaySubmitTestLine,
    displayNormalTestLine: rootState.assessmentStatisticsSiteState.displayNormalTestLine,
    displayManualPointsLine: rootState.assessmentStatisticsSiteState.displayManualPointsLine,

    langId: rootState.i18nState.langId

  }
}

const mapDispatchToProps = (dispatch: Dispatch) => bindActionCreators({
                                                                        //imported reducer funcs here
                                                                        setDisplaySubmitTestLine,
                                                                        setDisplayNormalTestLine,
                                                                        setDisplayManualPointsLine,
                                                                      }, dispatch)


const stateProps = returntypeof(mapStateToProps);
const dispatchProps = returntypeof(mapDispatchToProps);
type Props = typeof stateProps & typeof dispatchProps;


const normalPointsKey = ''
const submitPointsKey = ''

export interface AssessmentDiagramDataPoint {
  /**
   * some id to find the element (for tooltip)
   * use getAssessmentChartDataId
   */
  readonly id: string
  readonly displayName: string

  readonly originalSubmitPointsPercentageValue: number
  readonly originalNormalPointsPercentageValue: number

  /**
   * if should not count is true then we set it to 0 else originalSubmitPointsPercentageValue
   * we use NaN so that the diagram displays no point but we get a tooltip (entry)
   */
  readonly submitPointsPercentageValue: number
  /**
   * if should not count is true then we set it to 0 else originalNormalPointsPercentageValue
   * we use NaN so that the diagram displays no point but we get a tooltip (entry)
   */
  readonly normalPointsPercentageValue: number

  /**
   * we ignore should not count here (because the tutor assigned these points)
   * if we got more than 100% we clamp to 100%
   */
  readonly manualPointsPercentageValue: number

  /**
   * assessment result should not count
   */
  readonly shouldNotCount: boolean
}


export function getAssessmentChartDataId(point: SingleAssessmentStatistic): string {
  return `${point.exerciseId}-${point.exerciseReleaseId}-${point.pLangId}`
}

/**
 * this is (luckily) interpreted by rechart as none-value so we don't get a point
 *
 * so this is not assessed
 *
 * BUT DON'T COMPARE WITH === use isNaN !!!
 * (what is good if we haven assessed)
 * @type {number}
 */
export const defaultsToPercentageValue = Number.NaN

//--- just to ensure we use some right strings..
const getXAxisKey = (name: keyof AssessmentDiagramDataPoint) => name
const getNormalPointsPercentageKey = (name: keyof AssessmentDiagramDataPoint) => name
const getSubmitPointsPercentageKey = (name: keyof AssessmentDiagramDataPoint) => name
const getManualPointsPercentageKey = (name: keyof AssessmentDiagramDataPoint) => name


function getPercentage(points: number | null, maxPoints: number | null, defaultsTo: number): number {

  if (points === null || maxPoints === null) {
    return defaultsTo
  }

  //this is the case if we have not assessed this...
  if (points === 0 && maxPoints === 0) {
    // noinspection TsLint
    return defaultsToPercentageValue
  }

  return +(points * 100 / maxPoints).toFixed(2)
}


class ChartView extends React.Component<Props, any> {
  render(): JSX.Element {

    const list: ReadonlyArray<SingleAssessmentStatistic> = this.props.assessmentStatisticsList


    const graphData: ReadonlyArray<AssessmentDiagramDataPoint> = list.map(p => {

      const originalSubmitPointsPercentageValue = getPercentage(p.submitTestPoints, p.maxSubmitTestPoints,
                                                                defaultsToPercentageValue
      )

      const originalNormalPointsPercentageValue = getPercentage(p.normalTestPoints, p.maxNormalTestPoints,
                                                                defaultsToPercentageValue
      )

      const originalManualPointsPercentageValue = getPercentage(p.manualPoints, p.maxManualPoints,
                                                                defaultsToPercentageValue
      )

      //if not assessed (nan) then show nan
      //else only display value if should not count is false
      const submitPointsPercentageValue =
        isNaN(originalSubmitPointsPercentageValue)
        ? originalSubmitPointsPercentageValue
        : (p.shouldNotCount
           ? 0
           : originalSubmitPointsPercentageValue)

      const normalPointsPercentageValue =
        isNaN(originalNormalPointsPercentageValue)
        ? originalNormalPointsPercentageValue
        : (p.shouldNotCount
           ? 0
           : originalNormalPointsPercentageValue)


      const manualPointsPercentageValue = isNaN(originalManualPointsPercentageValue)
                                          ? originalManualPointsPercentageValue
                                          : Math.min(originalManualPointsPercentageValue, 100)

      return {
        id: getAssessmentChartDataId(p),
        displayName: displayDate(p.lastUpdatedAt, -1),

        originalSubmitPointsPercentageValue: originalSubmitPointsPercentageValue,
        originalNormalPointsPercentageValue: originalNormalPointsPercentageValue,

        submitPointsPercentageValue: submitPointsPercentageValue,
        normalPointsPercentageValue: normalPointsPercentageValue,
        shouldNotCount: p.shouldNotCount,
        manualPointsPercentageValue: manualPointsPercentageValue
      }
    })

    //the width of one line segment (dash array)
    const linePartSize = 5
    //the space between the normal, submit tests line
    const lineSpaceSize = 5

    // tslint:disable-next-line
    let linesCount = (this.props.displayNormalTestLine
                      ? 1
                      : 0)
      + (this.props.displaySubmitTestLine
         ? 1
         : 0)
      + (this.props.displayManualPointsLine
         ? 1
         : 0)
    const areMultipleLinesDisplayed = linesCount > 1

    // noinspection TsLint
    return (
      <SinglePanelSiteWrapper>

        {
          this.props.isLoading &&
          <Spinner/>
        }

        <div className="pad-right">
          <div className="view-head-line">
            <div className="view-caption">
              <h1>
                {
                  this.props.assessmentStatistics &&
                  `${this.props.assessmentStatistics.lastName}, ${this.props.assessmentStatistics.firstName}`
                }
              </h1>

            </div>
            <div className="view-options">
              <HelpPopup
                defaultText={getI18n(this.props.langId, "The graph only includes data displayed in the table")}/>
            </div>
          </div>

          <div className="view-content">

            <div className="mar-left-neg">
              <ResponsiveContainer
                width={'100%'}
                height={200}
                debounce={1000}
              >
                <LineChart
                  data={graphData as any[]}
                  width={500}
                  height={300}
                >
                  <XAxis dataKey={getXAxisKey('displayName')}/>
                  <YAxis unit='%'/>
                  <CartesianGrid stroke="#eee" strokeDasharray="5 5"/>
                  {
                    //payload is only for internal usage https://github.com/recharts/recharts/issues/530
                  }
                  {
                    (this.props.displaySubmitTestLine || this.props.displayNormalTestLine || this.props.displayManualPointsLine) &&
                    <Tooltip content={<AssessmentTooltip allRawData={this.props.assessmentStatisticsList}
                                                         displaySubmitTestLine={this.props.displaySubmitTestLine}
                                                         displayNormalTestLine={this.props.displayNormalTestLine}
                                                         displayManualPointsLine={this.props.displayManualPointsLine}
                                                         langId={this.props.langId}

                    />}/>
                  }

                  {
                    //not customizable (the labels)...
                    //<Legend/>
                  }

                  {
                    //current typings are invalid ... expecting string but docs and react says: number
                    //animationDuration={500}
                    //type="linear" is more accurate but looks uglier
                    //type="monotone" is not accurate if we have
                    //0% --> 50% --> 70% vs
                    //0% --> 50% --> 60%
                    //the the curves wont match between 0% and 50%...
                    /*

                      displaySubmitTestLine: rootState.assessmentStatisticsSiteState.displaySubmitTestLine,
                      displayNormalTestLine: rootState.assessmentStatisticsSiteState.displayNormalTestLine,
                     */
                  }


                  {
                    this.props.displayNormalTestLine &&
                    <Line type="monotone" dataKey={getNormalPointsPercentageKey('normalPointsPercentageValue')}
                          stroke={blueColor} strokeWidth={'2'}
                          isAnimationActive={false}
                          strokeDasharray={areMultipleLinesDisplayed
                                           ? `${linePartSize} ${lineSpaceSize * 2 + linePartSize}`
                                           : '0'}
                          strokeDashoffset={areMultipleLinesDisplayed
                                            ? `0`
                                            : '0'}

                    />
                  }

                  {
                    this.props.displaySubmitTestLine &&
                    <Line type="monotone" dataKey={getSubmitPointsPercentageKey('submitPointsPercentageValue')}
                          stroke={errorColor} strokeWidth={'2'}
                          isAnimationActive={false}
                          strokeDasharray={areMultipleLinesDisplayed
                                           ? `${linePartSize} ${lineSpaceSize * 2 + linePartSize}`
                                           : '0'}
                          strokeDashoffset={areMultipleLinesDisplayed
                                            ? `${linePartSize}`
                                            : '0'}

                    />
                  }



                  {
                    this.props.displayManualPointsLine &&
                    <Line type="monotone" dataKey={getManualPointsPercentageKey('manualPointsPercentageValue')}
                          stroke={greenColor} strokeWidth={'2'}
                          isAnimationActive={false}
                          strokeDasharray={areMultipleLinesDisplayed
                                           ? `${linePartSize} ${lineSpaceSize * 2 + linePartSize}`
                                           : '0'}
                          strokeDashoffset={areMultipleLinesDisplayed
                                            ? `${linePartSize*2}`
                                            : '0'}

                    />
                  }


                </LineChart>
              </ResponsiveContainer>
            </div>
            {
              //legend
            }
            <div className="ui form divided-checkboxes mar-left">
              <Form.Field>
                {
                  //submit points
                }
                <Checkbox label={getI18n(this.props.langId, 'Submit test points')}
                          checked={this.props.displaySubmitTestLine}
                          onChange={(e, data) => {
                            this.props.setDisplaySubmitTestLine(data.checked)
                          }}
                          className="mar-right-half"

                />

                {
                  //copied from react rechart
                }
                <svg className="recharts-surface legend-item mar-right-2x" width="14" height="14" viewBox="0 0 32 32"
                     version="1.1">
                  <path strokeWidth="4" fill="none" stroke={errorColor}
                        d="M0,16h10.666666666666666 A5.333333333333333,5.333333333333333,0,1,1,21.333333333333332,16 H32M21.333333333333332,16 A5.333333333333333,5.333333333333333,0,1,1,10.666666666666666,16"
                        className="recharts-legend-icon"></path>
                </svg>


                {
                  //normal points
                }
                <Checkbox label={getI18n(this.props.langId, 'Normal test points')}
                          checked={this.props.displayNormalTestLine}
                          onChange={(e, data) => {
                            this.props.setDisplayNormalTestLine(data.checked)
                          }}
                          className="mar-right-half"
                />
                <svg className="recharts-surface legend-item mar-right" width="14" height="14" viewBox="0 0 32 32"
                     version="1.1">
                  <path strokeWidth="4" fill="none" stroke={blueColor}
                        d="M0,16h10.666666666666666 A5.333333333333333,5.333333333333333,0,1,1,21.333333333333332,16 H32M21.333333333333332,16 A5.333333333333333,5.333333333333333,0,1,1,10.666666666666666,16"
                        className="recharts-legend-icon"></path>
                </svg>

                {
                  //manual points
                }
                <Checkbox label={getI18n(this.props.langId, 'Manual points')}
                          checked={this.props.displayManualPointsLine}
                          onChange={(e, data) => {
                            this.props.setDisplayManualPointsLine(data.checked)
                          }}
                          className="mar-right-half"
                />
                <svg className="recharts-surface legend-item mar-right" width="14" height="14" viewBox="0 0 32 32"
                     version="1.1">
                  <path strokeWidth="4" fill="none" stroke={greenColor}
                        d="M0,16h10.666666666666666 A5.333333333333333,5.333333333333333,0,1,1,21.333333333333332,16 H32M21.333333333333332,16 A5.333333333333333,5.333333333333333,0,1,1,10.666666666666666,16"
                        className="recharts-legend-icon"></path>
                </svg>


              </Form.Field>
            </div>

          </div>
        </div>


      </SinglePanelSiteWrapper>
    )
  }
}

export default connect(mapStateToProps, mapDispatchToProps)(ChartView)
