/**
 * Created by janis dähne.
 */

import * as React from "react";
import {connect} from "react-redux";
import {bindActionCreators, Dispatch} from "redux";
import {returntypeof} from 'react-redux-typescript';
import {RootState} from "../../../../state/reducers";
import {Button, Icon, List, Table} from "semantic-ui-react";
import {AssessedSubmitTestFromBackend, TestProtocolItem, TestProtocolType} from "../../../../types/testResults";
import ConsoleItem from './consoleItem'
import {convertTestResultToProtocolItems} from "../../../../helpers/convertersAndTransformers";
import {RunnableCustomTestFrontendExclusive} from "../../../../types/customTests";
import {KnownInternalTestTypes, TestTypeFullBase} from "../../../../types/testTypes";
import {HelpPopup} from "../../../helpers/helpPopup";
import {getI18n} from "../../../../../i18n/i18nRoot";
import {MyPopup} from '../../../helpers/myPopup'
import {setHideErrorOutputLines, setHideExpectedOutputLines} from '../../../../state/actions/globalUiActions'
import Logger from '../../../../helpers/logger'
import {RunnableCustomProjectTestFrontendExclusive} from '../../../../types/customProjects'
import {RunnableExerciseTestFrontendExclusive} from '../../../../types/exerciseEditor'
import {copyToClipboard} from '../../../../helpers/clipboard';

//const css = require('./styles.styl');

//only test OR protocolItems is valid
export interface MyProps {
  //readonly test: string
  readonly test: RunnableExerciseTestFrontendExclusive | RunnableCustomTestFrontendExclusive | AssessedSubmitTestFromBackend | RunnableCustomProjectTestFrontendExclusive | null
  readonly testTypes: ReadonlyArray<TestTypeFullBase>

  readonly protocolItems: ReadonlyArray<TestProtocolItem> | null

  /**
   * true: normal
   * false: hide the options to hide expected output & error output
   */
  readonly displayFilterConsoleItemsOptions: boolean

  /**
   * functions to load the full console output
   * we don't pass the test here because the test might have different types e.g. normal/custom test with have different funcs
   *   parent can decide better...
   */
  readonly loadFullOutputFunc: (() => Promise<void>) | null
}

const mapStateToProps = (rootState: RootState, props: MyProps) => {
  return {
    //test0: rootState...
    //test: props.test
    ...props,
    useDiffIfOutputMismatched: rootState.globalUiState.useDiffIfOutputMismatched,
    useReplaceWhiteSpacesWithBullets: rootState.globalUiState.useReplaceWhiteSpacesWithBullets,
    hideExpectedOutputLines: rootState.globalUiState.hideExpectedOutputLines,
    hideErrorOutputLines: rootState.globalUiState.hideErrorOutputLines,
    langId: rootState.i18nState.langId
  }
}

const mapDispatchToProps = (dispatch: Dispatch) => bindActionCreators({
                                                                        //imported reducer funcs here
                                                                        setHideExpectedOutputLines,
                                                                        setHideErrorOutputLines,
                                                                      }, dispatch)


const stateProps = returntypeof(mapStateToProps)
const dispatchProps = returntypeof(mapDispatchToProps);
type Props = typeof stateProps & typeof dispatchProps;

class ConsolePanelView extends React.Component<Props, any> {

  componentWillReceiveProps(nextProps: Readonly<Props>, nextContext: any): void {

    //display the stderr if the test didn't pass and the return code was !== 0 (error)
    //for expected return codes !== 0 we check if test passed
    if (this.props.test !== nextProps.test && nextProps.test && nextProps.test.result) {

      if (nextProps.test.result.passed === false && nextProps.test.result.programExitCode !== 0
        || nextProps.test.result.hasCompiled === false) {
        this.props.setHideErrorOutputLines(false)
      }
    }
  }

  copyTestResultInputToClipboard(): void {
    let protocolItemsCopy = this.getProtocolItemsSortedCopy()

    const inputString = protocolItemsCopy
      .filter(p => p.type === TestProtocolType.isInput)
      .map(p => p.content)
      .join('\n')
      .toString()

    copyToClipboard(inputString)
  }
  copyTestResultOutputToClipboard(): void {
    //the current visible state is copied
    //e.g. if debug output and if expected output
    let outputString = ""

    //TODO compare test result handling might be incorrect...

    //we need to get the test protocol items sorted because some e.g. expected return code not guaranteed to be the first item
    let protocolItemsCopy = this.getProtocolItemsSortedCopy()
      .filter(p => p.type === TestProtocolType.isOutput
        || p.type === TestProtocolType.isError
        || p.type === TestProtocolType.isExpectedOutput
        || p.type === TestProtocolType.isCompareTestResultOk
        || p.type === TestProtocolType.isCompareTestResultFile1NotFound
        || p.type === TestProtocolType.isCompareTestResultFile2NotFound
        || p.type === TestProtocolType.isCompareTestResultExpected
        || p.type === TestProtocolType.isCompareTestResultActual
        || p.type === TestProtocolType.isCompareTestResultFile2EOF
      )

    for (let i = 0; i < protocolItemsCopy.length; i++) {
      let protocolItem = protocolItemsCopy[i]

      if (this.props.hideExpectedOutputLines && (
        protocolItem.type === TestProtocolType.isExpectedOutput
        || protocolItem.type === TestProtocolType.isCompareTestResultExpected
        )
      ) {
        continue
      }

      if (this.props.hideErrorOutputLines && (
        protocolItem.type === TestProtocolType.isError
        || protocolItem.type === TestProtocolType.isCompareTestResultFile1NotFound
        || protocolItem.type === TestProtocolType.isCompareTestResultFile2NotFound
        || protocolItem.type === TestProtocolType.isCompareTestResultFile2EOF
        )
      ) {
        continue
      }

      outputString += protocolItem.content + '\n'
    }

    copyToClipboard(outputString)
  }
  copyTestResultInputAndOutputToClipboard(): void {
    //the current visible state is copied
    //e.g. if debug output and if expected output
    let outputString = ""

    //we need to get the test protocol items sorted because some e.g. expected return code not guaranteed to be the first item
    let protocolItemsCopy = this.getProtocolItemsSortedCopy()
      .filter(p => p.type === TestProtocolType.isInput
        || p.type === TestProtocolType.isOutput
        || p.type === TestProtocolType.isError
        || p.type === TestProtocolType.isExpectedOutput
        || p.type === TestProtocolType.isCompareTestResultOk
        || p.type === TestProtocolType.isCompareTestResultFile1NotFound
        || p.type === TestProtocolType.isCompareTestResultFile2NotFound
        || p.type === TestProtocolType.isCompareTestResultExpected
        || p.type === TestProtocolType.isCompareTestResultActual
        || p.type === TestProtocolType.isCompareTestResultFile2EOF
      )

    for (let i = 0; i < protocolItemsCopy.length; i++) {
      let protocolItem = protocolItemsCopy[i]

      if (this.props.hideExpectedOutputLines && (
        protocolItem.type === TestProtocolType.isExpectedOutput
        || protocolItem.type === TestProtocolType.isCompareTestResultExpected
      )
      ) {
        continue
      }

      if (this.props.hideErrorOutputLines && (
        protocolItem.type === TestProtocolType.isError
        || protocolItem.type === TestProtocolType.isCompareTestResultFile1NotFound
        || protocolItem.type === TestProtocolType.isCompareTestResultFile2NotFound
        || protocolItem.type === TestProtocolType.isCompareTestResultFile2EOF
      )
      ) {
        continue
      }

      outputString += protocolItem.content + '\n'
    }

    copyToClipboard(outputString)

  }

  //copy to not mutate...
  getProtocolItemsSortedCopy(): ReadonlyArray<TestProtocolItem> {

    const allTestTypes = this.props.testTypes.concat([{
      id: -1,
      displayName: 'Just Run Program',
      internalName: KnownInternalTestTypes.justRunTest
    }])

    // noinspection TsLint
    let protocolItems: ReadonlyArray<TestProtocolItem> = []

    if (this.props.test) {
      protocolItems = convertTestResultToProtocolItems(this.props.test, allTestTypes)
      // Logger.debug('consolePanelView', 'using test as content')
    }

    if (this.props.protocolItems) {
      protocolItems = this.props.protocolItems
      // Logger.debug('consolePanelView', 'using protocol items as content')
    }

    if (!protocolItems) {
      return []
    }

    //make a copy because we sort/mutate
    //make sure exit code is always 1.
    //expected always 2.
    //args 3.
    //because this looks better
    // tslint:disable-next-line
    let protocolItemsSorted = JSON.parse(JSON.stringify(protocolItems)) as TestProtocolItem[]

    //make sure exit code is first then expected exit code (we actually don't use them but we find isExpectedExitCode faster)
    // program args should come after those 2
    protocolItemsSorted.sort((a, b) => {

      if (a.type === TestProtocolType.isExpectedExitCode) {

        return -1
      }

      if (a.type === TestProtocolType.isProgramArguments) {

        if (b.type === TestProtocolType.isExpectedExitCode) return 1

        return -1
      }

      return 0

    })
    return protocolItemsSorted as ReadonlyArray<TestProtocolItem>
  }

  render(): JSX.Element {

    if (!this.props.test && !this.props.protocolItems) {
      //if we clear the console view we set both to null...
      //Logger.error('console panel view can only use test or protocolItems property')
      return null
    }

    //don't add this in the mapStateToProps else we already concat a new instance thus the shallow compare from redux is false
    //because we already create a new entry (with the same values but other identity)
    const allTestTypes = this.props.testTypes.concat([{
      id: -1,
      displayName: 'Just Run Program',
      internalName: KnownInternalTestTypes.justRunTest
    }])

    //only allow if we have expected output lines
    //maybe this is too strange because you cannot toggle the item and then one might forget to uncheck it...
    const canHideExpectedOutputLines = true //protocolItems.some(p => p.type === TestProtocolType.isExpectedOutput)
    const canHideErrorOutputLines = true //protocolItems.some(p => p.type === TestProtocolType.isError)


    //maybe not sort here but search for the items?? but we could have more than one...
    //performance could be better

    //make sure exit code is always 1.
    //expected always 2.
    //args 3.
    //because this looks better
    // tslint:disable-next-line
    let protocolItemsSorted = this.getProtocolItemsSortedCopy()

    // tslint:disable-next-line
    let expectedExitCode: number | null = null
    // tslint:disable-next-line
    let exitCode: number | null = this.props.test && this.props.test.result
                                  ? this.props.test.result.programExitCode
                                  : null

    const expectedExitCodeItem = protocolItemsSorted.find(p => p.type === TestProtocolType.isExpectedExitCode)

    if (expectedExitCodeItem) {
      expectedExitCode = parseInt(expectedExitCodeItem.content, 10)

      if (isNaN(expectedExitCode)) {
        Logger.warn(
          `expectedExitCode '${expectedExitCodeItem.content}' could not be converted to int`)
        // shouldDisplayExitCode = false; //user should see the error...
      }
    }

    return (
      <div className="fh">

        <List divided relaxed>
          {
            //only display header if we have some normal output... (or just don't display headers if all items are errors)
            (protocolItemsSorted && protocolItemsSorted.some(p => p.type !== TestProtocolType.isError)) &&
            <div className="console-input-output-line-header">
              <div className="input-part">
                <div className="flexed-well-spaced">

                  <div className="flexed">
                    <Icon className="clickable" name="clipboard outline" onClick={() => this.copyTestResultInputToClipboard()} />
                    <div className="mar-right-half">
                      {
                        getI18n(this.props.langId, 'Input')
                      }
                    </div>
                    {
                      this.props.useReplaceWhiteSpacesWithBullets &&
                      <HelpPopup className="warning-colored"
                                 defaultText={getI18n(this.props.langId,
                                                      'All whitespaces of the input are replaced with · and all tabs with ⇥'
                                 )}/>
                    }

                  </div>

                  <Icon className="not-important-colored" name="chevron right"/>
                </div>
              </div>
              <div className="program-part-invisible">

                <div className="absolute-centered flexed">

                  <div className="mar-right-half">
                    <Icon className="clickable" name="clipboard outline" onClick={() => this.copyTestResultInputAndOutputToClipboard()} />
                  </div>

                  {
                    canHideErrorOutputLines && this.props.displayFilterConsoleItemsOptions &&
                    <MyPopup
                      trigger={
                        <div className={`${this.props.hideErrorOutputLines
                                           ? 'clickable-toggled'
                                           : 'clickable'}`}
                             onClick={() => {
                               this.props.setHideErrorOutputLines(!this.props.hideErrorOutputLines)
                             }}
                        >
                          {
                            //TODO looks a bit ugly...
                          }
                          <div className="fa-slashed-div">
                            <Icon name='bug'/>
                            <div className={`fa-slash ${this.props.hideErrorOutputLines
                                                        ? ' toggle-active'
                                                        : ''}`}></div>
                          </div>

                        </div>
                      }
                      content={
                        <div>
                          {getI18n(this.props.langId,
                                   "If this is enabled all error output lines (stderr) will be hidden. Keep in mind that this will also hide any real error message"
                          )}
                        </div>
                      }
                      wide={'very'}
                    />
                  }

                  <div className="mar-right-half mar-left-half">
                    {
                      getI18n(this.props.langId, 'Program')
                    }
                  </div>

                  {
                    canHideExpectedOutputLines && this.props.displayFilterConsoleItemsOptions &&
                    <MyPopup
                      trigger={
                        <div className={`${this.props.hideExpectedOutputLines
                                           ? 'clickable-toggled'
                                           : 'clickable'}`}
                             onClick={() => {
                               this.props.setHideExpectedOutputLines(!this.props.hideExpectedOutputLines)
                             }}
                        >
                          <Icon name="hide"/>
                        </div>
                      }
                      content={
                        <div>
                          {getI18n(this.props.langId,
                                   "If this is enabled all expected output lines will be hidden and only lines from your program are displayed"
                          )}
                        </div>
                      }
                      wide
                    />
                  }

                </div>

              </div>
              <div className="output-part">
                <div className="flexed-well-spaced">
                  <Icon className="not-important-colored" name="chevron right"/>

                  <div className="flexed">
                    <Icon className="clickable" name="clipboard outline" onClick={() => this.copyTestResultOutputToClipboard()} />
                    <div className="mar-right-half">
                      {
                        getI18n(this.props.langId, 'Output')
                      }
                    </div>
                    {
                      this.props.useReplaceWhiteSpacesWithBullets &&
                      <HelpPopup className="warning-colored"
                                 defaultText={getI18n(this.props.langId,
                                                      'All whitespaces of the output are replaced with · and all tabs with ⇥'
                                 )}/>
                    }

                  </div>
                </div>
              </div>
            </div>
          }

          {
            this.props.test && this.props.test.result &&
            <List.Item style={{padding: 0}}>
              <div className="time-measure-line console-arguments-line">

                {
                  //display test-server and test-runner version
                }
                <MyPopup
                  trigger={
                    <Icon className="hoverable mar-right-half" name="stopwatch"/>
                  }
                  content={
                    <Table sortable selectable padded striped singleLine className="material-table">
                      <Table.Header>
                        <Table.Row>
                          <Table.HeaderCell>
                            {getI18n(this.props.langId, "Property")}
                          </Table.HeaderCell>
                          <Table.HeaderCell>
                            {getI18n(this.props.langId, "Value")}
                          </Table.HeaderCell>
                        </Table.Row>
                      </Table.Header>
                      <Table.Body>

                        <Table.Row>
                          <Table.Cell>
                            {getI18n(this.props.langId, "Used timeout in ms")}
                          </Table.Cell>
                          <Table.Cell>
                            {
                              this.props.test.result.timeoutInMsUsed
                            }
                          </Table.Cell>
                        </Table.Row>

                        <Table.Row>
                          <Table.Cell>
                            {getI18n(this.props.langId, "Used compile timeout in ms")}
                          </Table.Cell>
                          <Table.Cell>
                            {
                              this.props.test.result.compileTimeoutInMsUsed
                            }
                          </Table.Cell>
                        </Table.Row>

                        <Table.Row>
                        <Table.Cell>
                          {getI18n(this.props.langId, "Used character limit")}
                        </Table.Cell>
                        <Table.Cell>
                          {
                            this.props.test.result.characterLimitUsed
                          }
                        </Table.Cell>
                      </Table.Row>

                        <Table.Row>
                          <Table.Cell>
                            {getI18n(this.props.langId, "Test-server version")}
                          </Table.Cell>
                          <Table.Cell>
                            {
                              this.props.test.result.testServerVersion
                            }
                          </Table.Cell>
                        </Table.Row>

                        <Table.Row>
                          <Table.Cell>
                            {getI18n(this.props.langId, "Test-runner version")}
                          </Table.Cell>
                          <Table.Cell>
                            {
                              this.props.test.result.runnerVersion
                            }
                          </Table.Cell>
                        </Table.Row>
                      </Table.Body>
                    </Table>
                  }
                />


                <span>{getI18n(this.props.langId, 'compiled in')}: </span>
                {
                  this.props.test.result.timeForCompiling
                  ?
                  <span>
                    <span>{this.props.test.result.timeForCompiling} </span>
                    <span>{getI18n(this.props.langId, "ms")}</span>
                  </span>
                  : <span>-</span>
                }

                {
                  this.props.test.result.timeForUserProgram !== null &&
                  <span>
                      <span>, {getI18n(this.props.langId, "executed in")}: </span>
                      <span>{this.props.test.result.timeForUserProgram} </span>
                      <span>{getI18n(this.props.langId, "ms")}</span>
                    </span>
                }


              </div>
            </List.Item>
          }

          {
            (this.props.test
              && allTestTypes.find(
                p => this.props.test.testTypeId === p.id).internalName === KnownInternalTestTypes.compileTest) &&
            this.props.test.result && this.props.test.result.hasCompiled &&
            <List.Item style={{padding: 0}}>
              <div className="ok-colored console-arguments-line">
                {
                  getI18n(this.props.langId, 'Compiled')
                }
              </div>
            </List.Item>
          }
          {
            ((exitCode !== null && exitCode !== 0) || expectedExitCode !== null) &&
            <List.Item style={{padding: 0}}>
              <div className="flexed console-arguments-line">
                <div className="mar-right-half">
                  {getI18n(this.props.langId, 'Program exit code')}:
                </div>
                <div className="flexed">
                  {
                    expectedExitCode === null &&
                    <div>
                      <span className="error-colored">{exitCode}</span>

                      <HelpPopup className="mar-left-half" defaultText={getI18n(this.props.langId,
                                                                                "A normal program has an exit code of 0. If you intended an exit code different from 0 then specify this in the test"
                      )}
                                 wide
                      />
                    </div>
                  }

                  {
                    expectedExitCode !== null && expectedExitCode === exitCode &&
                    <span className="ok-colored">{exitCode}</span>
                  }

                  {
                    expectedExitCode !== null && expectedExitCode !== exitCode &&
                    <div>
                      <span className="error-colored">{exitCode}</span>
                      <span className="error-colored mar-left-half">
                      <span>({getI18n(this.props.langId, "but expected exit code")} </span>
                      <span>{expectedExitCode})</span>
                    </span>
                    </div>
                  }

                </div>
              </div>
            </List.Item>
          }
          {
            protocolItemsSorted &&
            protocolItemsSorted.map((item: TestProtocolItem, index) => {

                                      const nextItem = index + 1 >= protocolItemsSorted.length
                                                       ? null
                                                       : protocolItemsSorted[index + 1]

                                      const prevItem = index - 1 < 0
                                                       ? null
                                                       : protocolItemsSorted[index - 1]

                                      if (canHideExpectedOutputLines && this.props.hideExpectedOutputLines && item.type === TestProtocolType.isExpectedOutput) {
                                        return null
                                      }

                                      if (canHideErrorOutputLines && this.props.hideErrorOutputLines && item.type === TestProtocolType.isError) {
                                        return null
                                      }

                                      return (
                                        <ConsoleItem key={index}
                                                     item={item}
                                                     nextItem={nextItem}
                                                     prevItem={prevItem}
                                                     useDiffIfOutputMismatched={this.props.useDiffIfOutputMismatched}
                                                     useReplaceWhiteSpacesWithBullets={this.props.useReplaceWhiteSpacesWithBullets}
                                        />
                                      )
                                    }
            )
          }

          {
            (this.props.test !== null && this.props.test.result !== null && this.props.test.result.hasMoreTestProtocolLines) && this.props.loadFullOutputFunc &&
              <Button fluid onClick={() => {

                if (!this.props.loadFullOutputFunc) return

                this.props.loadFullOutputFunc()
              }}>{getI18n(this.props.langId, 'Load full output')}</Button>
          }
        </List>
      </div>
    )
  }
}

const comp = connect(mapStateToProps, mapDispatchToProps)(ConsolePanelView)
export default comp
