/**
 * Created by janis dähne.
 */
import {isNotNullOrUndefined} from "../validation/validations";
import {
  EditorExerciseForBackend,
  EditorExerciseFromBackend,
  ExerciseFromBackendWithData,
  ExerciseTestFromBackend,
  ExerciseTestFromBackendWithData,
  RunnableExerciseTestFrontendExclusive
} from "../types/exerciseEditor";
import {Moment} from "moment";
import {DoExerciseFromBackend, DoExerciseFrontendExclusive} from "../types/doExercise";
import {
  CustomTestForBackend,
  CustomTestFromBackend,
  RunnableAfterCustomTestFrontendExclusive,
  RunnableCustomTestFrontendExclusive
} from "../types/customTests";
import {
  AssessedSubmitTestFromBackend,
  AssessedSubmitTestFrontendExclusive,
  StoredTestResultFromBackend,
  StoredTestResultTupleFromBackend,
  TestProtocolItem,
  TestProtocolType
} from "../types/testResults";
import {TestAnswerFromBackend, TestProtocol} from "../types/testsAndTestProtocol";
import {KnownInternalTestTypes, TestTypeFullBase} from "../types/testTypes";
import {startsWith} from "./stringHelper";
import Logger from "./logger";
import {ExercisePreviewsFrontendExclusive} from "../components/sites/allAccessibleExercisesSite/openExercises/allOpenExercisesSite";
import {ExercisePreviewFromBackend} from "../types/exercisePreview";
import {errorDialog} from "./dialogHelper";
import {
  CustomProjectFrontendExclusive,
  CustomProjectFullBase,
  CustomProjectTestFullBase,
  RunnableCustomProjectTestFrontendExclusive
} from "../types/customProjects";
import {orderBy} from 'lodash-es'
import {getI18n, KnownLangs} from "../../i18n/i18nRoot";
import globalState from '../state/state'
import {upState} from './typeHelpers'


//from http://stackoverflow.com/questions/15900485/correct-way-to-convert-size-in-bytes-to-kb-mb-gb-in-javascript
/**
 * converts the num of bytes to a human readable format
 * @param {number} bytes
 * @param {number} decimals num decimals
 * @returns {any}
 */
export function convertSizeToHumanReadableSize(bytes: number, decimals = 0): string {
  if (bytes === 0) return '0 Byte'
  const k = 1024
  const dm = decimals < 0 ? 0 : decimals;
  const sizes: ReadonlyArray<string> = ['Bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB']
  const i = Math.floor(Math.log(bytes) / Math.log(k))
  //parse float avaids trailing zeros
  return `${parseFloat((bytes / Math.pow(k, i)).toFixed(dm))} ${sizes[i]}`
}

/**
 * checks if the given id string is an id an returns it
 *
 * id is a number and > 0
 * @param {string} id
 * @returns {number | null}
 */
export function parseId(id: string): number | null {

  if (!isNotNullOrUndefined(id)) return null

  // noinspection TsLint
  const check = parseInt(id, 10)

  if (isNaN(check) || check <= 0) {
    return null
  }

  return check
}

export function parseNotNegativeOrNullNumber(id: string): number | null {

  if (!isNotNullOrUndefined(id)) return null

  // noinspection TsLint
  const check = parseInt(id, 10)

  if (isNaN(check) || check <= 0) {
    return null
  }

  return check
}

/**
 * parses a number
 * if the string is null or undefined or parsed NaN null is returned
 * @param id
 */
export function parseAnyId(id: string | null | undefined): number | null {
  if (!isNotNullOrUndefined(id)) return null

  // noinspection TsLint
  const check = parseInt(id, 10)

  if (isNaN(check)) {
    return null
  }

  return check
}

/**
 * a frontend id is an id < 0
 * @param {string} id
 * @returns {number}
 */
export function parseFrontendId(id: string): number | null {

  if (!isNotNullOrUndefined(id)) return null

  // noinspection TsLint
  const check = parseInt(id, 10)

  if (isNaN(check) || check >= 0) {
    return null
  }

  return check
}


export function parseBoolean(test: string): boolean | null {

  if (!isNotNullOrUndefined(test)) return null

  test = test.toLowerCase()

  if (typeof test === 'string') {

    if (test === 'true') {
      return true
    } else if (test === 'false') {
      return false
    }
  }

  return null
}


//userLangId could be used to display date/time for different countries
export function displayDateAndTime(dateAndTime: Moment, userLangId: number | -1): string {

  if (!dateAndTime) return ''

  return dateAndTime.format('HH:mm:ss - DD.MM.YYYY')
}

export function displayTime(dateAndTime: Moment, userLangId: number | -1): string {

  if (!dateAndTime) return ''

  return dateAndTime.format('HH:mm:ss')
}

export function displayDate(dateAndTime: Moment, userLangId: number | -1): string {

  if (!dateAndTime) return ''

  return dateAndTime.format('DD.MM.YYYY')
}

export function displayNumber(val: number, decimalPlaces: number): string {
  return val.toFixed(decimalPlaces)
}


/**
 * creates a flat copy of the exercise and converts it
 * @param {EditorExerciseFromBackend} exercise
 * @returns {EditorExerciseForBackend}
 */
export function convertEditorExerciseFromBackendToFrontend(exercise: EditorExerciseFromBackend): EditorExerciseForBackend {

  // backend now includes test content or sets it to null
  // const copyTests: ReadonlyArray<ExerciseTestForBackend> = exercise.tests.map(test => {
  //   const copy: ExerciseTestForBackend = {
  //     ...test,
  //     files: test.files.map((f: TestAssetFromBackendWithData) => {
  //
  //       const asset: TestAssetForBackend = {
  //         ...f,
  //         content: []
  //       }
  //
  //       return asset
  //     })
  //   }
  //   return copy
  // })

  const {...rest} = exercise

  return upState(exercise, {
    ...rest,
    creatorId: exercise.creatorId,
    tests: orderBy(exercise.tests, p => p.displayIndex)
  })
}

/**
 * merges testsWithContent into the exercise
 * DOES CREATE A REF TO testsWithContent
 * @param {EditorExerciseFromBackend} exercise
 * @param {ReadonlyArray<ExerciseTestFromBackend>} testsWithContent
 * @returns {EditorExerciseFromBackend}
 */
export function mergeEditorExerciseTestAssetContents(exercise: EditorExerciseFromBackend,
                                                     testsWithContent: ReadonlyArray<ExerciseTestFromBackendWithData>
): ExerciseFromBackendWithData {

  return {
    ...exercise,
    tests: testsWithContent
  }
}

/**
 *
 * use an extra function for this to easily refactor this
 * CREATES A COPY
 * @param {ReadonlyArray<AssessedSubmitTestFromBackend>} assessments
 * @param areAfterTests sets the ui is after test prop
 * @returns {ReadonlyArray<AssessedSubmitTestFrontendExclusive>}
 */
export function convertAssessmentFromBackendToFrontend(assessments: ReadonlyArray<AssessedSubmitTestFromBackend>,
                                                       areAfterTests: boolean
): ReadonlyArray<AssessedSubmitTestFrontendExclusive> {
  return assessments.map<AssessedSubmitTestFrontendExclusive>((p) => {
    return {
      ...p,
      uiIsRunning: false,
      uiIsAfterTest: areAfterTests,
      type: 'submit',
    }
  })
}

/**
 * CREATES A COPY
 * @param {ReadonlyArray<ExerciseTestFromBackend>} tests
 * @param areAfterTests sets the ui is after test prop
 * @returns {ReadonlyArray<RunnableExerciseTestFrontendExclusive>}
 */
export function convertTestFromBackendToFrontend(tests: ReadonlyArray<ExerciseTestFromBackend>,
                                                 areAfterTests: boolean
): ReadonlyArray<RunnableExerciseTestFrontendExclusive> {
  return tests.map<RunnableExerciseTestFrontendExclusive>((p) => {
    return {
      ...p,
      uiIsRunning: false,
      result: null,
      uiIsAfterTest: areAfterTests,
      type: 'normal',
    }
  })
}

/**
 * creates a copy of the exercise and converts it
 * @param {DoExerciseFromBackend} exercise
 * @returns {DoExerciseFrontendExclusive}
 */
export function convertDoExerciseFromBackendToFrontend(exercise: DoExerciseFromBackend): DoExerciseFrontendExclusive {

  const copy: DoExerciseFrontendExclusive = {
    ...exercise,
    tests: exercise.tests.map(test => {
      const runnableTest: RunnableExerciseTestFrontendExclusive = {
        ...test,
        uiIsRunning: false,
        result: null,
        uiIsAfterTest: false,
        files: test.files, // backend now includes test content or sets it to null
        //   test.files.map<TestAssetForBackend>(file => {
        //   const forBackend: TestAssetForBackend = {
        //     ...file,
        //     content: []
        //   }
        //   return forBackend
        // })
        type: 'normal',
      }
      return runnableTest
    }),

    customTests: exercise.customTests.map(test => {
      const runnableTest: RunnableCustomTestFrontendExclusive = {
        ...test,
        uiIsRunning: false,
        result: null,
        uiIsAfterTest: false,
        files: test.files, // backend now includes test content or sets it to null
        //   test.files.map<TestAssetForBackend>(file => {
        //   const forBackend: TestAssetForBackend = {
        //     ...file,
        //     content: []
        //   }
        //   return forBackend
        // })
        type: 'custom',
      }
      return runnableTest
    })
  }

  return copy
}

/**
 * creates a copy of the project and converts it
 * @param {DoExerciseFromBackend} project
 * @returns {DoExerciseFrontendExclusive}
 */
export function convertCustomProjectFromBackendToFrontend(project: CustomProjectFullBase): CustomProjectFrontendExclusive {

  const copy: CustomProjectFrontendExclusive = {
    ...project,
    tests: project.tests.map(test => {
      const runnableTest: RunnableCustomProjectTestFrontendExclusive = {
        ...test,
        uiIsRunning: false,
        result: null,
        uiIsAfterTest: false,
        files: test.files, // backend now includes test content or sets it to null
        type: 'customProjectTest',
      }
      return runnableTest
    }),
  }

  return copy
}

/**
 * merges the test results in the exercise (tests) and returns a new copy
 * @param {DoExerciseFrontendExclusive} exercise
 * @param {StoredTestResultTupleFromBackend} allTests
 * @returns {DoExerciseFrontendExclusive}
 */
export function mergeTestResultsIntoDoExercise(exercise: DoExerciseFrontendExclusive,
                                               allTests: StoredTestResultTupleFromBackend
): DoExerciseFrontendExclusive {

  const copy: DoExerciseFrontendExclusive = {
    ...exercise,
    tests: exercise.tests.map(test => {

      const storedResult = allTests.testResults.find(p => p?.testId === test.id)

      const runnableTest: RunnableExerciseTestFrontendExclusive = {
        ...test,
        uiIsRunning: false,
        result: _storedTestResultToTestResult(storedResult)
      }
      return runnableTest
    }),

    customTests: exercise.customTests.map(test => {

      const storedResult = allTests.customTestResults.find(p => p?.testId === test.id)

      const runnableTest: RunnableCustomTestFrontendExclusive = {
        ...test,
        uiIsRunning: false,
        result: _storedTestResultToTestResult(storedResult)
      }
      return runnableTest
    })
  }

  return copy
}

/**
 * copied from mergeTestResultsIntoDoExercise
 * merges the test results in the exercise (tests) and returns a new copy
 */
export function mergeTestResultsIntoCustomProject(exercise: CustomProjectFullBase,
                                                  allTests: ReadonlyArray<StoredTestResultFromBackend | null>
): CustomProjectFrontendExclusive {

  const copy: CustomProjectFrontendExclusive = {
    ...exercise,
    tests: exercise.tests.map(test => {

      const storedResult = allTests.find(p => p?.testId === test.id)

      const runnableTest: RunnableCustomProjectTestFrontendExclusive = {
        ...test,
        uiIsRunning: false,
        uiIsAfterTest: false,
        result: _storedTestResultToTestResult(storedResult),
        type: 'customProjectTest',
      }
      return runnableTest
    })
  }

  return copy
}

// https://github.com/Microsoft/TypeScript/issues/14409 --> https://github.com/Microsoft/TypeScript/issues/10727 --> typescript 2.7
// export function mergeNormalTestResults<T extends RunnableExerciseTestFrontendExclusive | RunnableCustomTestFrontendExclusive>(
//   tests: ReadonlyArray<T>, results: ReadonlyArray<TestAnswerFromBackend>): ReadonlyArray<T> {
//
//   return tests.map<T>(test => {
//
//     const testResult = results.find((p) => p.testId === test.id)
//
//     //do not change the test
//     if (!testResult) return {
//       ...test, // https://github.com/Microsoft/TypeScript/issues/14409 --> https://github.com/Microsoft/TypeScript/issues/10727 --> typescript 2.7
//       uiIsRunning: false
//     }
//
//     const copy: T = {
//       ...test,
//       result: testResult,
//       uiIsRunning: false,
//     }
//     return copy
//   })
// }

/**
 * returns A COPY
 * @param {ReadonlyArray<RunnableExerciseTestFrontendExclusive>} tests
 * @param {ReadonlyArray<StoredTestResultFromBackend>} results
 * @returns {ReadonlyArray<RunnableExerciseTestFrontendExclusive>}
 */
export function mergeStoredNormalTestResults(tests: ReadonlyArray<RunnableExerciseTestFrontendExclusive>,
                                             results: ReadonlyArray<StoredTestResultFromBackend>
): ReadonlyArray<RunnableExerciseTestFrontendExclusive> {

  return tests.map<RunnableExerciseTestFrontendExclusive>(test => {

    const testResult = results.find((p) => p.testId === test.id)

    //do not change the test
    if (!testResult) return {
      ...test, // https://github.com/Microsoft/TypeScript/issues/14409 --> https://github.com/Microsoft/TypeScript/issues/10727 --> typescript 2.7
      uiIsRunning: false
    }

    const copy: RunnableExerciseTestFrontendExclusive = {
      ...test,
      result: _storedTestResultToTestResult(testResult),
      uiIsRunning: false,
    }
    return copy
  })
}

/**
 * returns A COPY
 * @param {ReadonlyArray<RunnableExerciseTestFrontendExclusive>} tests
 * @param {ReadonlyArray<TestAnswerFromBackend>} results
 * @param resetTestResultsBefore true: resets the test results ... if the new test result is now found then no test result is placed,
 * false: do not clear the test result (so in case results is an empty array no results will be changed
 * @returns {ReadonlyArray<RunnableExerciseTestFrontendExclusive>}
 */
export function mergeNormalTestResults(tests: ReadonlyArray<RunnableExerciseTestFrontendExclusive>,
                                       results: ReadonlyArray<TestAnswerFromBackend>, resetTestResultsBefore = false
): ReadonlyArray<RunnableExerciseTestFrontendExclusive> {

  return tests.map<RunnableExerciseTestFrontendExclusive>(test => {

    const testResult = results.find((p) => p.testId === test.id)

    //do not change the test
    if (!testResult) return {
      ...test, // https://github.com/Microsoft/TypeScript/issues/14409 --> https://github.com/Microsoft/TypeScript/issues/10727 --> typescript 2.7
      result: resetTestResultsBefore
              ? null
              : test.result,
      uiIsRunning: false
    }

    const copy: RunnableExerciseTestFrontendExclusive = {
      ...test,
      result: testResult,
      uiIsRunning: false,
    }
    return copy
  })
}

/**
 * returns A COPY
 * @param {ReadonlyArray<RunnableCustomTestFrontendExclusive>} tests
 * @param {ReadonlyArray<TestAnswerFromBackend>} results
 * @param resetTestResultsBefore true: resets the test results ... if the new test result is now found then no test result is placed,
 * false: do not clear the test result (so in case results is an empty array no results will be changed
 * @returns {ReadonlyArray<RunnableCustomTestFrontendExclusive>}
 */
export function mergeCustomTestResults(tests: ReadonlyArray<RunnableCustomTestFrontendExclusive>,
                                       results: ReadonlyArray<TestAnswerFromBackend>, resetTestResultsBefore = false
): ReadonlyArray<RunnableCustomTestFrontendExclusive> {

  return tests.map<RunnableCustomTestFrontendExclusive>(test => {

    const testResult = results.find((p) => p.testId === test.id)

    //do not change the test
    if (!testResult) return {
      ...test, // https://github.com/Microsoft/TypeScript/issues/14409 --> https://github.com/Microsoft/TypeScript/issues/10727 --> typescript 2.7
      result: resetTestResultsBefore
              ? null
              : test.result,
      uiIsRunning: false
    }

    const copy: RunnableCustomTestFrontendExclusive = {
      ...test,
      result: testResult,
      uiIsRunning: false,
    }
    return copy
  })
}

export function mergeAfterCustomTestResults(tests: ReadonlyArray<RunnableAfterCustomTestFrontendExclusive>,
                                            results: ReadonlyArray<TestAnswerFromBackend>,
                                            resetTestResultsBefore = false
): ReadonlyArray<RunnableAfterCustomTestFrontendExclusive> {

  return tests.map<RunnableAfterCustomTestFrontendExclusive>(test => {

    const testResult = results.find((p) => p.testId === test.id)

    //do not change the test
    if (!testResult) return {
      ...test, // https://github.com/Microsoft/TypeScript/issues/14409 --> https://github.com/Microsoft/TypeScript/issues/10727 --> typescript 2.7
      result: resetTestResultsBefore
              ? null
              : test.result,
      uiIsRunning: false
    }

    const copy: RunnableAfterCustomTestFrontendExclusive = {
      ...test,
      result: testResult,
      uiIsRunning: false,
    }
    return copy
  })
}

/**
 *
 * @param {ReadonlyArray<AssessedSubmitTestFrontendExclusive>} tests
 * @param {ReadonlyArray<TestAnswerFromBackend>} results
 * @param resetTestResultsBefore true: resets the test results ... if the new test result is now found then no test result is placed,
 * false: do not clear the test result (so in case results is an empty array no results will be changed
 * @returns {ReadonlyArray<AssessedSubmitTestFrontendExclusive>}
 */
export function mergeSubmitTestResults(tests: ReadonlyArray<AssessedSubmitTestFrontendExclusive>,
                                       results: ReadonlyArray<TestAnswerFromBackend>, resetTestResultsBefore = false
): ReadonlyArray<AssessedSubmitTestFrontendExclusive> {

  return tests.map<AssessedSubmitTestFrontendExclusive>(test => {

    const testResult = results.find((p) => p.testId === test.id)

    //do not change the test
    if (!testResult) return {
      ...test, // https://github.com/Microsoft/TypeScript/issues/14409 --> https://github.com/Microsoft/TypeScript/issues/10727 --> typescript 2.7
      result: resetTestResultsBefore
              ? null
              : test.result,
      uiIsRunning: false
    }

    const copy: AssessedSubmitTestFrontendExclusive = {
      ...test,
      result: testResult,
      uiIsRunning: false,
    }
    return copy
  })
}


/**
 * copied from mergeCustomTestResults
 * @param {ReadonlyArray<RunnableCustomProjectTestFrontendExclusive>} tests
 * @param {ReadonlyArray<TestAnswerFromBackend>} results
 * @param {boolean} resetTestResultsBefore
 * @returns {ReadonlyArray<RunnableCustomProjectTestFrontendExclusive>}
 */
export function mergeCustomTestResultsCustomProject(tests: ReadonlyArray<RunnableCustomProjectTestFrontendExclusive>,
                                                    results: ReadonlyArray<TestAnswerFromBackend>,
                                                    resetTestResultsBefore = false
): ReadonlyArray<RunnableCustomProjectTestFrontendExclusive> {

  return tests.map<RunnableCustomProjectTestFrontendExclusive>(test => {

    const testResult = results.find((p) => p.testId === test.id)

    //do not change the test
    if (!testResult) return {
      ...test, // https://github.com/Microsoft/TypeScript/issues/14409 --> https://github.com/Microsoft/TypeScript/issues/10727 --> typescript 2.7
      result: resetTestResultsBefore
              ? null
              : test.result,
      uiIsRunning: false
    }

    const copy: RunnableCustomProjectTestFrontendExclusive = {
      ...test,
      result: testResult,
      uiIsRunning: false,
    }
    return copy
  })
}

export function _storedTestResultToTestResult(storedResult: StoredTestResultFromBackend | null): TestAnswerFromBackend | null {

  if (!storedResult) return null

  return {
    hasCompiled: storedResult.hasCompiled,
    passed: storedResult.passed,
    points: storedResult.points,
    testId: storedResult.testId,
    programExitCode: storedResult.programExitCode,
    protocol: storedResult.protocol,
    testResultCode: storedResult.testResultCode,
    testServerMessage: null,
    testServerCode: null,
    runnerVersion: storedResult.runnerVersion,
    testServerVersion: storedResult.testServerVersion,
    timeForCompiling: storedResult.timeForCompiling,
    timeForUserProgram: storedResult.timeForUserProgram,
    timeoutInMsUsed: storedResult.timeoutInMsUsed,
    compileTimeoutInMsUsed: storedResult.compileTimeoutInMsUsed,
    characterLimitExceeded: storedResult.characterLimitExceeded,
    characterLimitUsed: storedResult.characterLimitUsed,
    hasMoreTestProtocolLines: storedResult.hasMoreTestProtocolLines,
  }
}


/**
 * converts the test result protocol items to proper TestProtocolItem
 * @param {RunnableExerciseTestFrontendExclusive | RunnableCustomTestFrontendExclusive} test
 * @param {ReadonlyArray<TestTypeFullBase>} testTypes
 * @param ignoreEmptyLines ignore empty ('') protocol lines,
 *    actually the test server ignores every test content line that starts with an unknown char...
 * @returns {ReadonlyArray<TestProtocolItem>}
 */
export function convertTestResultToProtocolItems(test: RunnableExerciseTestFrontendExclusive | RunnableCustomTestFrontendExclusive | AssessedSubmitTestFromBackend | RunnableCustomProjectTestFrontendExclusive,
                                                 testTypes: ReadonlyArray<TestTypeFullBase>,
                                                 ignoreEmptyLines = true
): ReadonlyArray<TestProtocolItem> | null {

  if (!test || !test.result) {

    errorDialog(getI18n(globalState.getState().i18nState.langId, 'Error'),
                getI18n(globalState.getState().i18nState.langId, 'Internal error (see console)'),
                globalState.getState().i18nState.langId
    )
    Logger.error(`no test or test result to convert to protocol`)

    return null
  }

  if (!testTypes || testTypes.length === 0) {

    errorDialog(getI18n(globalState.getState().i18nState.langId, 'Error'),
                getI18n(globalState.getState().i18nState.langId, 'Internal error (see console)'),
                globalState.getState().i18nState.langId
    )

    Logger.error(`no test types to convert to protocol`)

    return null
  }

  const testType = testTypes.find(p => p.id === test.testTypeId)

  if (!testType) {

    errorDialog(getI18n(globalState.getState().i18nState.langId, 'Error'),
                getI18n(globalState.getState().i18nState.langId, 'Unknown test type'),
                globalState.getState().i18nState.langId
    )
    Logger.error(`could not find test type for id ${test.testTypeId}`)

    return null
  }

  if (test.result.hasCompiled === false) {
    //display errors
    return convertNotCompiledMessages(test, testType)
  }

  //compiled ... convert
  return convertTestProtocol(test.result.protocol, testType)
}


//TODO make split more tough?
export function convertTestProtocolFromString(testContent: string,
                                              testType: TestTypeFullBase
): ReadonlyArray<TestProtocolItem> | null {
  return convertTestProtocol(testContent.split('\n'), testType)
}

export function convertTestProtocol(protocol: ReadonlyArray<string>,
                                    testType: TestTypeFullBase,
                                    ignoreEmptyLines = true
): ReadonlyArray<TestProtocolItem> | null {


  // do not transform the lines here e.g. trim or replace spaces...
  // better do this only when displaying (so we can access still the original string)

  if (!testType) return null

  if (testType.internalName === KnownInternalTestTypes.blackBoxTest || testType.internalName === KnownInternalTestTypes.justRunTest) {

    //this is better to process via

    // noinspection TsLint
    let items: TestProtocolItem[] = []

    const protocolItems = protocol
      .filter(line => line !== '')
      .map<TestProtocolItem>(line => {

        line = trimStart(line)  //for some reason some lines start with whitespaces...

         if (startsWith(line, TestProtocol.argumentSign)) {
          return {
            type: TestProtocolType.isProgramArguments,
            hasOutputMismatched: false,
            content: line.substring(TestProtocol.argumentSign.length),
          }
        } else if (startsWith(line, TestProtocol.outputSign)) {

          //assume correct item (hasOutputMismatched = false)
          //if the next line is an expected output line then this
          //line was wrong...
          //check this afterwards
          return {
            type: TestProtocolType.isOutput,
            hasOutputMismatched: false,
            content: line.substring(TestProtocol.outputSign.length),
          }

        } else if (startsWith(line, TestProtocol.expectedPrefix)) {
          return {
            type: TestProtocolType.isExpectedOutput,
            hasOutputMismatched: false,
            content: line.substring(TestProtocol.expectedPrefix.length),
          }
        } else if (startsWith(line, TestProtocol.inputSign)) {
          return {
            type: TestProtocolType.isInput,
            hasOutputMismatched: false,
            content: line.substring(TestProtocol.inputSign.length),
          }
        } else if (startsWith(line, TestProtocol.errorPrefix)) {
          return {
            type: TestProtocolType.isError,
            hasOutputMismatched: false,
            content: line.substring(TestProtocol.errorPrefix.length),
          }
        } else if (line === TestProtocol.endOfUserProgramString) {
          return {
            type: TestProtocolType.isEndOfTestLine,
            hasOutputMismatched: false,
            content: TestProtocol.endOfUserProgramString,
          }
          // this is the last because only e would also match expected and vice versa
        } else if (line === TestProtocol.missingOutput) {
           return {
             type: TestProtocolType.isMissingOutput,
             hasOutputMismatched: false,
             content: '',
           }
        } else if (line.length > 0 && line[0] === TestProtocol.regexOutputSign) {

          return {
            type: TestProtocolType.isOutput,
            hasOutputMismatched: false,
            //the test server returns (regex): if the output didn't match
            content: `(regex):${line.substring(TestProtocol.regexOutputSign.length)}`,
          }
        }
        else if (startsWith(line, TestProtocol.expectedProgramExitCodePrefix)) { //error, expected also start with e
          return {
            type: TestProtocolType.isExpectedExitCode,
            hasOutputMismatched: false,
            content: line.substring(TestProtocol.expectedProgramExitCodePrefix.length),
          }
        }


        //could be some debug output...

        return {
          type: TestProtocolType.unknown,
          hasOutputMismatched: false,
          content: line,
        }
      })

    //now set all hasOutputMismatched (if an output item is followed by an isExpectedOutput item)

    const protocolItemsFinished = protocolItems.map<TestProtocolItem>((item, index) => {

      if (item.type === TestProtocolType.isOutput) {
        //this is an output item
        //check next item

        if (index + 1 >= protocolItems.length) {
          //this was the last item ... all ok with this item
          return item
        }

        //we have a next item
        if (protocolItems[index + 1].type === TestProtocolType.isExpectedOutput) {
          //the next item describes an expectedOutput so the output mismatched
          return {
            ...item,
            hasOutputMismatched: true
          }
        }
      }

      return item
    })

    return protocolItemsFinished

  }
  else if (testType.internalName === KnownInternalTestTypes.compileTest) {

    //we only wanted to know if it's compiles or not...
    //but there could be some debug lines or std error (e.g. from linters)...
    return protocol.map<TestProtocolItem>(line => {

      line = trimStart(line)  //for some reason some lines start with whitespaces...

      if (startsWith(line, TestProtocol.errorPrefix)) {
        return {
          type: TestProtocolType.isError,
          hasOutputMismatched: false,
          content: line.substring(TestProtocol.errorPrefix.length),
        }
      }

      return {
        type: TestProtocolType.unknown,
        hasOutputMismatched: false,
        content: line,
      }
    })

  }
  else if (testType.internalName === KnownInternalTestTypes.externalCheckerTest || testType.internalName === KnownInternalTestTypes.externalCheckerStructuralInductionTest) {

    return protocol.map<TestProtocolItem>(line => {

      line = trimStart(line)  //for some reason some lines start with whitespaces...

      if (startsWith(line, TestProtocol.outputSign)) {

        //some normal lines from the checker
        return {
          type: TestProtocolType.plain,
          hasOutputMismatched: false,
          content: line.substring(TestProtocol.outputSign.length),
        }

      }
      else if (startsWith(line, TestProtocol.errorPrefix)) {
        return {
          type: TestProtocolType.isError,
          hasOutputMismatched: false,
          content: line.substring(TestProtocol.errorPrefix.length),
        }
      }


      return {
        type: TestProtocolType.unknown,
        hasOutputMismatched: false,
        content: line,
      }
    })

  }
  else if (testType.internalName === KnownInternalTestTypes.regexTest) {

    //use plain output
    return protocol.map<TestProtocolItem>(line => {
      return {
        type: TestProtocolType.plain,
        hasOutputMismatched: false,
        content: line,
      }
    })
  }
  else if (testType.internalName === KnownInternalTestTypes.compareFilesTest) {


    const protocolItems = protocol
      .filter(line => line !== '')
      .map<TestProtocolItem>( (line) => {

        line = trimStart(line) //for some reason some lines start with whitespaces...

         if (startsWith(line, TestProtocol.argumentSign)) {
          return {
            type: TestProtocolType.isProgramArguments,
            hasOutputMismatched: false,
            content: line.substring(TestProtocol.argumentSign.length),
          }
        } else if (line === TestProtocol.endOfUserProgramString) {
          return {
            type: TestProtocolType.isEndOfTestLine,
            hasOutputMismatched: false,
            content: TestProtocol.endOfUserProgramString,
          }

        } //real compare output lines
        else if (line.startsWith(TestProtocol.compareOutputFileTestInput)) {
          return {
            type: TestProtocolType.isCompareTestCompare,
            hasOutputMismatched: false,
            content: line,
          }

        } else if (line.startsWith(TestProtocol.compareOutputFileTestOutputOk)) {
          return {
            type: TestProtocolType.isCompareTestResultOk,
            hasOutputMismatched: false,
            content: line,
          }
        }
        else if (line.startsWith(TestProtocol.compareOutputFileTestOutputActual)) {
          return {
            type: TestProtocolType.isCompareTestResultActual,
            hasOutputMismatched: false,
            content: line,
          }
        }
        else if (line.startsWith(TestProtocol.compareOutputFileTestOutputExpected)) {
          return {
            type: TestProtocolType.isCompareTestResultExpected,
            hasOutputMismatched: false,
            content: line,
          }
        }
        else if (line.startsWith(TestProtocol.compareOutputFileTestOutputFile1NotFound)) {
          return {
            type: TestProtocolType.isCompareTestResultFile1NotFound,
            hasOutputMismatched: false,
            content: line,
          }
        }
        else if (line.startsWith(TestProtocol.compareOutputFileTestOutputFile2NotFound)) {
          return {
            type: TestProtocolType.isCompareTestResultFile2NotFound,
            hasOutputMismatched: false,
            content: line,
          }
        }

        //could be some debug output...

        return {
          type: TestProtocolType.unknown,
          hasOutputMismatched: false,
          content: line,
        }
      })


    return protocolItems

  }
  else {
    //TODO unknown test type
    Logger.warn("found unknown test protocol type (internal name)")
  }

  return null
}

/**
 * converts the test protocol lines into compiler error items
 * @param {RunnableExerciseTestFrontendExclusive | RunnableCustomTestFrontendExclusive} test
 * @param {TestTypeFullBase} testType
 * @returns {ReadonlyArray<TestProtocolItem>}
 */
function convertNotCompiledMessages(test: RunnableExerciseTestFrontendExclusive | RunnableCustomTestFrontendExclusive | AssessedSubmitTestFromBackend | RunnableCustomProjectTestFrontendExclusive,
                                    testType: TestTypeFullBase
): ReadonlyArray<TestProtocolItem> {

  return test.result.protocol.map<TestProtocolItem>(line => {

    line = trimStart(line)

    if (line.startsWith(TestProtocol.errorPrefix)) {
      return {
        type: TestProtocolType.isError,
        hasOutputMismatched: false,
        content: line.substring(TestProtocol.errorPrefix.length),
      }
    }

    return {
      type: TestProtocolType.unknown,
      hasOutputMismatched: false,
      content: line,
    }
  })
}

/**
 * converts a CustomTestFromBackend to a CustomTestForBackend (content will be empty)
 * @param {CustomTestFromBackend} test
 * @returns {CustomTestForBackend}
 */
export function convertCustomTestFromBackendToFrontend(test: CustomTestFromBackend): CustomTestForBackend {

  const copy: CustomTestForBackend = {
    ...test,
    files: test.files
  }

  return copy
}

export function convertCustomTestFromBackendToRunnableFrontendCustomTest(test: CustomTestForBackend,
                                                                         areAfterTests: boolean
): RunnableCustomTestFrontendExclusive {

  const copy: RunnableCustomTestFrontendExclusive = {
    ...test,
    uiIsRunning: false,
    result: null,
    uiIsAfterTest: areAfterTests,
    files: test.files,
    type: 'custom',
  }
  return copy
}

export function convertCustomProjectTestFromBackendToRunnableFrontendCustomProjectTest(test: CustomProjectTestFullBase): RunnableCustomProjectTestFrontendExclusive {

  const copy: RunnableCustomProjectTestFrontendExclusive = {
    ...test,
    uiIsRunning: false,
    result: null,
    uiIsAfterTest: false,
    files: test.files,
    type: 'customProjectTest',
  }
  return copy
}

export function convertOpenExercisePreviewToFrontendOpenExercisePreview(exercisePreview: ReadonlyArray<ExercisePreviewFromBackend>): ReadonlyArray<ExercisePreviewsFrontendExclusive> {


  const copy = exercisePreview.map<ExercisePreviewsFrontendExclusive>(p => {
    return {
      ...p,
      leftTimeString: '',
      timeIsUp: false,
      isStillOpen: false,
      isDisplayed: true,
      leftTimeInS: null
    }
  })

  return copy
}

function trimStart(text: string): string {

  // tslint:disable-next-line
  let index = 0

  while (text[index] === ' ') {
    index++

  }

  return text.substring(index)

}

/**
 * converts the assessment result to a string
 * @param langId
 * @param hasAutomaticAssessmentFinished
 * @param assessmentIsEmpty {@link AssessmentFullBase.hasAssessment} ?
 * @param testPoints
 * @param maxTestPoints
 * @param ignoreHasAssessmentFinished true: param {@link hasAutomaticAssessmentFinished} is ignored, false: not
 */
export function assessmentPointsToString(langId: KnownLangs, hasAutomaticAssessmentFinished: boolean, assessmentIsEmpty: boolean, testPoints: number | null, maxTestPoints: number | null, ignoreHasAssessmentFinished: boolean): string {

  if (assessmentIsEmpty) {
    return getI18n(langId, 'Not assessed')
  }

  if (ignoreHasAssessmentFinished === false && hasAutomaticAssessmentFinished === false) {
    return getI18n(langId, 'Not assessed')
  }

  return `${isNotNullOrUndefined(testPoints) 
            ? testPoints 
            : 'null'} / ${isNotNullOrUndefined(maxTestPoints)
                 ? maxTestPoints
                 : 'null'}`

}
