/**
 * Created by janis dähne.
 */

import {Moment} from "moment";
import * as moment from "moment";
import {isProduction} from '../../debug'

export enum SeverityLevel {
  info,
  warning,
  error,
  debug
}

/**
 * a log item from the logger
 */
export interface LogItem {
  readonly id: number
  readonly file: string | null
  readonly level: SeverityLevel
  readonly message: string
  readonly params?: ReadonlyArray<any>
  readonly time?: Moment
  /**
   * true: do not display in ui (only blank item), only when exporting
   */
  readonly isSilent: boolean
}


/**
 * a common logger
 */
export default class Logger {

  static nextItemId = 1

  // noinspection TsLint
  static myLog: LogItem[] = []

  // noinspection TsLint
  static log(message?: any, ...optionalParams: any[]): void {
    Logger.logWithLevel(message, SeverityLevel.info, null, false, optionalParams)
  }

  // noinspection TsLint
  static warn(message?: any, ...optionalParams: any[]): void {
    Logger.logWithLevel(message, SeverityLevel.warning, null, false, optionalParams)
  }

  // noinspection TsLint
  static error(message?: any, ...optionalParams: any[]): void {

    if (isProduction) {
      Logger.logWithLevel(message, SeverityLevel.error, null, false, optionalParams)
    } else {
      console.warn(message) //error would block error dialogs...
      Logger.logWithLevel(message, SeverityLevel.error, null, false, optionalParams)
    }

  }

  // noinspection TsLint
  static todo(message?: any, ...optionalParams: any[]): void {
    Logger.logWithLevel(message, SeverityLevel.debug, null, false, optionalParams)
  }

  // noinspection TsLint
  /**
   * makes a debug output formatted as '[place] message'
   * @param place the location (file)
   * @param message
   * @param optionalParams
   */
  static debug(place: string, message: any, ...optionalParams: any[]): void {
    Logger.logWithLevel(message, SeverityLevel.debug, place, false, optionalParams)
  }

  // noinspection TsLint
  /**
   * same as debug but not displayed in ui... this is useful for some sensitive logging
   * @param place
   * @param message
   * @param optionalParams
   */
  static debugSilent(place: string, message: any, ...optionalParams: any[]): void {
    Logger.logWithLevel(message, SeverityLevel.debug, place, true, optionalParams)
  }

  // noinspection TsLint
  private static logWithLevel(message: string, level: SeverityLevel, file: string | null,
                              isSilent: boolean,
                              ...optionalParams: any[]
  ): void {

    if (!isSilent) {
      if (file) {
        console.log(`[${file}] ${message}`, optionalParams)
      } else {
        console.log(message, optionalParams)
      }
    }

    //we cannot enable this because if we encountered an error during mapStateToProps
    //this gets called with dispatches
    //causing every (mounted) mapStateToProps to run again...

    Logger.myLog.push({
                        id: Logger.nextItemId++,
                        message,
                        level,
                        file,
                        params: optionalParams,
                        time: moment(),
                        isSilent
                      })

    //TODO find workaround
    // store.dispatch(addLogItem({
    //   id: Logger.nextItemId++,
    //   message,
    //   level,
    //   file,
    //   params: optionalParams,
    //   time: moment()
    // }))

  }

}
