import axios, {AxiosResponse} from 'axios'
import {dateTimeReviverLocalTime} from "./helpers/jsonTransformer";
import globalState from './state/state'
import store from './state/state'
//import before... to be sure??? see https://craig.is/killing/mice --> plugins
import {setUserData} from "./state/actions/settingsSite/userSettingsSiteActions";
import {userDataInitial} from "./state/reducers/settingsSite/settingsSiteReducers";
import {
  supportMail,
  crsfCookieName,
  getCookie,
  refreshServerMessageIntervalInS,
  setCurrentCrsfToken,
  versionString,
} from "./constants";
import {ServerResponse, ServerResponseWithData} from "./types/response";
import {ResponseCodes} from "./helpers/responseCodes";

import {isProduction} from "../debug";
import {getI18n} from "../i18n/i18nRoot";
import Logger from './helpers/logger'
import {geGlobalServerMessageAsync,} from './state/actions/globalSystemSettings/globalSystemSettingsCrudActions'
import {Error} from 'tslint/lib/error'
import {setBeforeLoginErrorMessage} from './state/actions/loginActions'
import {setLocalServerMessage} from './state/actions/globalSystemSettings/globalSystemSettingsActions'
import {ErrorHelper} from './helpers/errorHelper'
import {errorDialog} from './helpers/dialogHelper'
import {getServerMessageUrl} from './communicationLayer/systemSettingsLayer'

declare var require: any

require('react-datetime/css/react-datetime.css')
//require('../customChangedLib/react-notifications/notifications.styl') //imported via allStylues.styl

//TODO https://bjacobel.com/2016/12/04/highlight-bundle-size/
//hightlight js
require('highlight.js/styles/atom-one-light.css')
require('github-markdown-css/github-markdown.css')

require('sweetalert2/dist/sweetalert2.min.css')

require('mousetrap/plugins/global-bind/mousetrap-global-bind.js')
require('mousetrap/plugins/pause/mousetrap-pause.js')

const hljs = require('highlight.js/lib/highlight')
//https://bjacobel.com/2016/12/04/highlight-bundle-size/
//import * as hljs from 'highlight.js';

/**
 * various setup steps for the app
 *
 * this should be called before everything else
 */
export async function setup(): Promise<void> {

  Logger.log('setup started...')
  Logger.log('frontend version is ' + versionString)

  //https://developer.mozilla.org/de/docs/Web/API/GlobalEventHandlers/onerror
  window.onerror = function (message: any, file: string, line: number, col: number, error: Error): boolean {
    Logger.warn(`caught unhandled error!! ${JSON.stringify(message)};${file};${line};${col};${JSON.stringify(error)}`)
    return false //throw error
  };


  setInterval(async () => {

    try {
      await globalState.dispatch(geGlobalServerMessageAsync())
      Logger.debug('setup.ts', 'updated system settings successfully')
    } catch (err) {
      //also possible if not yet logged in
      Logger.debug('setup.ts', 'could not get refresh system settings')
    }


  }, refreshServerMessageIntervalInS * 1000)

  //axios.defaults.responseType = 'json'

  //https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/withCredentials
  //we want cross-origin requests to include cookies for develop...
  if (!isProduction) {
    axios.defaults.withCredentials = true
  }

  //check if we already have a token... (could also be from another tab because cookie is shared)
  //we now get the csrf token initially (cookies stored for all tabs) so if we have already the csrf token the first request to check login will not fail because of the csrf token the auth token is already sent as cookie so we don't need to get it here
  const csrfCookie = getCookie(crsfCookieName)
  setCurrentCrsfToken(csrfCookie)
  axios.defaults.headers['X-CSRF-Token'] = csrfCookie;

  axios.defaults.transformResponse = (response: string) => {
    //this removes the default json transformer and we will get the plain string
    return response
  }

  // axios.defaults.headers = {
  //   ...axios.defaults.headers,
  //   post: {
  //     'content-type': 'application/json'
  //   }
  // }

  //logged in? interceptor
  axios.interceptors.response.use((response) => {

                                    const csrfCookie = getCookie(crsfCookieName)
                                    setCurrentCrsfToken(csrfCookie)
                                    axios.defaults.headers['X-CSRF-Token'] = csrfCookie;

                                    return response
                                  },
                                  (error: {readonly response: AxiosResponse}) => {

                                    if (error && (error as any).response) {
                                      const response = error.response as any
                                      //if we use a reverse proxy (e..g nginx) we get a status code of 502...
                                      if (response && (typeof response.data === 'string' || response.data instanceof String) //the response is normally the html error site...
                                        && response.status === 502
                                      ) {
                                        // console.log('11111111111')
                                        setNotOnlineStatusMessage(true)
                                      }
                                    } else if (error && error.response === undefined) {  //this is for develop/no reverse proxy...
                                      const response = error as any
                                      if (!response.data && !response.status) { //see https://github.com/axios/axios/issues/383 --> should be network error
                                        // console.log('2222222222222')

                                        if (response.config.url.endsWith(getServerMessageUrl)) {
                                            //do not show a dialog because this is called every x mins...
                                          setNotOnlineStatusMessage(false)
                                        }
                                        else {
                                          setNotOnlineStatusMessage(true)
                                        }

                                      }
                                    }

                                    if (!error.response) {
                                      Logger.error(error)
                                      return error
                                    }

                                    // noinspection TsLint
                                    if (error && error.response && error.response.status === 401) {
                                      //reset user data to show login

                                      Logger.log("unauthorized... resetting user data to show login")
                                      store.dispatch(setUserData(userDataInitial))
                                    }

                                    // noinspection TsLint
                                    if (error && error.response && error.response.status === 403) { //no/wrong csrf-token
                                      //reset user data to show login

                                      Logger.log("invalid csrf... resetting user data to show login")
                                      store.dispatch(setUserData(userDataInitial))
                                    }

                                    return Promise.reject(error.response)
                                  }
  )


  axios.interceptors.response.use((response: AxiosResponse) => {

    if (response.headers && (!response.headers['content-type'] || response.headers['content-type'] === 'application/json')) {
      if (response.data) {


        try {
          if (globalState.getState().loginDataState.beforeLoginErrorMessage !== '') {
            globalState.dispatch(setBeforeLoginErrorMessage(''))
          }

          if (globalState.getState().globalSystemSettingsState.localServerMessage !== '') {
            globalState.dispatch(setLocalServerMessage('', false))
          }
        } catch (err) {
          //just swallow the error... maybe we read or set during some reducer steps
        }

        //this will convert all string dates into moment objects
        const obj: ServerResponse | ServerResponseWithData<any> = JSON.parse(response.data, dateTimeReviverLocalTime)

        //server errors are normal json objects with some special codes

        if (obj.code !== ResponseCodes.ok && obj.code !== ResponseCodes.partialOk) {
          //some error/warning
          Logger.error(`setup caught not OK response code: ${obj.code}, message: ${obj.message}`)

          //this is displayed on every error
          //but this is executed first so it's hidden behind the more specific error dialogs
          ErrorHelper.makeServerErrorDialog({
                                              ...response,
                                              data: obj
                                            }, globalState.getState().i18nState.langId, 'Server error', null)

          //reject
          //return Promise.reject(error.response)
        }

        response.data = obj
      }
    }
    else if (response.headers['content-type'] !== 'application/json') {
      //do not modify the response...
    }


    return response
  })


  Logger.log('...setup finished')


}

function setNotOnlineStatusMessage(displayErrorDialog: boolean): void {

  globalState.dispatch(setBeforeLoginErrorMessage(`${getI18n(globalState.getState().i18nState.langId,
                                                             "Could not connect to the server (the server might be offline), probably a server maintenance is the reason. If the problem persists, contact:"
  )} ${supportMail}`))

  globalState.dispatch(setLocalServerMessage(`${getI18n(globalState.getState().i18nState.langId,
                                                        "Could not connect to the server (the server might be offline), probably a server maintenance is the reason. You can keep this tab open (e.g. if you have unsaved changes) and try it later again. Every action that requires connecting to the server will do the state check. If the problem persists, contact:"
  )} ${supportMail}`, true))

  if (displayErrorDialog) {

    errorDialog(getI18n(globalState.getState().i18nState.langId, "Error"),
                `${getI18n(globalState.getState().i18nState.langId,
                           "Could not connect to the server (the server might be offline), probably a server maintenance is the reason. You can keep this tab open (e.g. if you have unsaved changes) and try it later again. Every action that requires connecting to the server will do the state check. If the problem persists, contact:"
                )} ${supportMail}`, globalState.getState().i18nState.langId
    )
  }

  //when we try to get the server message every 1 minute this popup is probably bad...
  // errorDialog('', 'Could not connect to the server (the server might be offline), probably a server maintenance is the reason. You can keep this tab open (e.g. if you have unsaved changes) and try it later again. Every action that requires connecting to the server will do the state check', globalState.getState().i18nState.langId)

  //set as not logged in ... this is bad because if the user has unsaved changes...
  // globalState.dispatch(resetUserSettingsSite())
}
