/**
 * Created by janis dähne.
 */


import {ActionBase, ActionType} from "./changeCodeTemplateDialogActionTypes";
import {ExerciseCodeTemplateFullBase} from "../../../../../types/exerciseEditor";
import {
  AllActions as CodeTemplateActions,
  initial as initialCodeTemplate,
  reducer as codeTemplateReducer, validationRules
} from './codeTemplateReducer'
import {isValidDialog} from "../../../../../validation/dialogValidations";
import {notExhaustive} from "../../../_notExhausiveHelper";
import {HasFileChangedMap} from "../../../../../helpers/editorHelper";
import {ActionTypeGlobal, RESET_GlobalAction} from '../../../globalActions'
import {AceEditorsStorageHelper} from '../../../../../components/codeEditors/aceEditors/aceEditorsStorageHelper'
import {editExerciseChangeCodeTemplateViewEditorId} from '../../../../../constants'

export type State = {

  readonly codeTemplate: ExerciseCodeTemplateFullBase

  readonly selectedEditorTabId: number

  readonly isCreating: boolean
  readonly isValid: boolean
  readonly isBusy: boolean

  readonly isMonacoFileExplorerDisplayed: boolean

  /**
   * file name for new files
   */
  readonly changeCodeTemplateNewFileName: string

  /**
   * true: display file renaming
   */
  readonly isRenamingCodeTemplateFile: boolean

  /**
   * the new file name for binding to the input
   */
  readonly renamingCodeTemplateFileName: string

  readonly renamingFileIdCodeTemplate: number

  readonly hasUnsavedChanges: boolean,

  /**
   * we don't use this because performance is too bad ... the whole editor rerenders every time?
   */
  readonly hasFileChangedMap: HasFileChangedMap


}

/*
 * the ace storage is only synced once when we load the exercise!
 */
export const initial: () => State = () => {
  return {
    codeTemplate: initialCodeTemplate(),
    isValid: isValidDialog(initialCodeTemplate(), validationRules),
    selectedEditorTabId: -1,
    isCreating: false,
    isBusy: false,
    isMonacoFileExplorerDisplayed: true,
    changeCodeTemplateNewFileName: '',
    isRenamingCodeTemplateFile: false,
    renamingCodeTemplateFileName: '',
    renamingFileIdCodeTemplate: NaN, //use nan because a new file coud have id -1

    hasUnsavedChanges: false,
    hasFileChangedMap: {}

  }
}


export interface SET_hasFileChangedMapAction extends ActionBase {
  readonly type: ActionType.SET_hasFileChangedMap
  readonly hasFileChangedMap: HasFileChangedMap
}

export interface SET_hasUnsavedChangesAction extends ActionBase {
  readonly type: ActionType.SET_hasUnsavedChanges
  readonly hasUnsavedChanges: boolean
}

export interface SET_codeTemplateAction extends ActionBase {
  readonly type: ActionType.SET_codeTemplate
  readonly codeTemplate: ExerciseCodeTemplateFullBase
}


export interface SET_isCreatingAction extends ActionBase {
  readonly type: ActionType.SET_isCreating
  readonly isCreating: boolean
}

export interface SET_selectedEditorTabIdAction extends ActionBase {
  readonly type: ActionType.SET_selectedEditorTabId
  readonly selectedEditorTabId: number
}


export interface SET_isMonacoFileExplorerDisplayedAction extends ActionBase {
  readonly type: ActionType.SET_isMonacoFileExplorerDisplayed
  readonly isMonacoFileExplorerDisplayed: boolean
}

export interface SET_changeCodeTemplateNewFileNameAction extends ActionBase {
  readonly type: ActionType.SET_changeCodeTemplateNewFileName
  readonly changeCodeTemplateNewFileName: string
}

export interface SET_isRenamingCodeTemplateFileAction extends ActionBase {
  readonly type: ActionType.SET_isRenamingCodeTemplateFile
  readonly isRenamingCodeTemplateFile: boolean
}

export interface SET_renamingCodeTemplateFileNameAction extends ActionBase {
  readonly type: ActionType.SET_renamingCodeTemplateFileName
  readonly renamingCodeTemplateFileName: string
}

export interface SET_renamingFileIdCodeTemplateAction extends ActionBase {
  readonly type: ActionType.SET_renamingFileIdCodeTemplate
  readonly renamingFileIdCodeTemplate: number
}


export interface ResetAction extends ActionBase {
  readonly type: ActionType.RESET
}

export type AllActions =
  ResetAction

  | SET_codeTemplateAction
  | SET_selectedEditorTabIdAction
  | SET_changeCodeTemplateNewFileNameAction
  | SET_isCreatingAction
  | SET_hasUnsavedChangesAction

  | SET_isMonacoFileExplorerDisplayedAction
  | SET_isRenamingCodeTemplateFileAction
  | SET_renamingCodeTemplateFileNameAction
  | SET_renamingFileIdCodeTemplateAction
  | SET_hasFileChangedMapAction

  | CodeTemplateActions

  | RESET_GlobalAction


export function reducer(state: State = initial(), action: AllActions): State {

  switch (action.type) {

    case ActionType.SET_codeTemplate: {

      const tabs = AceEditorsStorageHelper.getEditor(editExerciseChangeCodeTemplateViewEditorId)
      if (tabs) {
        //we want to reset the content but also want to keep the undo stack...
        //poping undo elements it probably now that good so we set the original content of the code template file
        //that will push a new undo item but that is ok
        for (let i = 0; i < tabs.length; i++) {
          const tabWrapper = tabs[i]

          const codeTemplateFile = action.codeTemplate.templateFiles.find(p => p.id === tabWrapper.tab.id)

          if (!codeTemplateFile) continue

          tabWrapper.tab = {
            ...tabWrapper.tab,
            id: codeTemplateFile.id,
            content: codeTemplateFile.content,
          }
          tabWrapper.session.setValue(codeTemplateFile.content)
        }
      }

      return {
        ...state,
        isValid: isValidDialog(action.codeTemplate, validationRules),
        codeTemplate: action.codeTemplate,
      }
    }

    case ActionType.SET_pLangId:
    case ActionType.SET_mainFileId:
    case ActionType.SET_templateFiles:

      const copy = codeTemplateReducer(state.codeTemplate, action)

      return {
        ...state,
        isValid: isValidDialog(copy, validationRules),
        codeTemplate: copy,
      }

    case ActionType.SET_hasUnsavedChanges:
      return {
        ...state,
        hasUnsavedChanges: action.hasUnsavedChanges
      }

    case ActionType.SET_selectedEditorTabId:
      return {
        ...state,
        selectedEditorTabId: action.selectedEditorTabId
      }

    case ActionType.SET_isMonacoFileExplorerDisplayed:
      return {
        ...state,
        isMonacoFileExplorerDisplayed: action.isMonacoFileExplorerDisplayed
      }

    case ActionType.SET_changeCodeTemplateNewFileName:
      return {
        ...state,
        changeCodeTemplateNewFileName: action.changeCodeTemplateNewFileName
      }

    case ActionType.SET_isCreating:
      return {
        ...state,
        isCreating: action.isCreating,
        isValid: isValidDialog(state.codeTemplate, validationRules) //we need this because for creating we only call reset
      }

    case ActionType.SET_renamingCodeTemplateFileName:
      return {
        ...state,
        renamingCodeTemplateFileName: action.renamingCodeTemplateFileName
      }

    case ActionType.SET_isRenamingCodeTemplateFile:
      return {
        ...state,
        isRenamingCodeTemplateFile: action.isRenamingCodeTemplateFile
      }

    case ActionType.SET_renamingFileIdCodeTemplate:
      return {
        ...state,
        renamingFileIdCodeTemplate: action.renamingFileIdCodeTemplate
      }

    case ActionType.SET_hasFileChangedMap:
      return {
        ...state,
        hasFileChangedMap: action.hasFileChangedMap
      }


    case ActionType.RESET:
    case ActionTypeGlobal.RESET_Global:
      return initial()

    default:
      notExhaustive(action)
      return state
  }
}

