import {Action} from "redux";
import {notExhaustive} from './_notExhausiveHelper'
import {
  defaultTaskDescriptionFontSizeInPx
} from '../../constants'
import {FrontendSettingsManager} from '../../helpers/frontendSettingsManager'
import {ActionTypeGlobal, RESET_GlobalAction} from './globalActions'

export type State = {
  readonly taskDescriptionFontSizeInPx: number

  /**
   * this will overwrite the setting in the user settings for editor font size (because this will be reset on reload)
   * if null use the user settings
   */
  readonly quickCodeEditorFontSizeInPx: number | null


  /**
   * for the console view
   * true: uses diff instead of real output
   */
  readonly useDiffIfOutputMismatched: boolean

  /**
   * for the console view
   * true: replaces whitespaces with bullets and tabs with -> utf-8 char
   */
  readonly useReplaceWhiteSpacesWithBullets: boolean

  /**
   * true: hide the output output lines
   * false: normal, display expected output lines
   */
  readonly hideExpectedOutputLines: boolean

  /**
   * true: hide the error output lines
   * false: normal, display them
   */
  readonly hideErrorOutputLines: boolean
}


export const initial: State = {
  taskDescriptionFontSizeInPx: defaultTaskDescriptionFontSizeInPx,
  quickCodeEditorFontSizeInPx: null,

  useDiffIfOutputMismatched: FrontendSettingsManager.getFrontendSettings().useDiffIfOutputMismatched,
  useReplaceWhiteSpacesWithBullets: FrontendSettingsManager.getFrontendSettings().useReplaceWhiteSpacesWithBullets,

  hideExpectedOutputLines: false,
  hideErrorOutputLines: false,
}

export interface ActionBase extends Action {
  readonly type: ActionType
}


export enum ActionType {
  SET_taskDescriptionFontSizeInPx = 'globalUiReducer_SET_taskDescriptionFontSizeInPx',
  SET_quickCodeEditorFontSizeInPx = 'globalUiReducer_SET_quickCodeEditorFontSizeInPx',

  SET_useDiffIfOutputMismatched = 'globalUiReducer_SET_useDiffIfOutputMismatched',
  SET_useReplaceWhiteSpacesWithBullets = 'globalUiReducer_SET_useReplaceWhiteSpacesWithBullets',

  SET_hideExpectedOutputLines = 'globalUiReducer_SET_hideExpectedOutputLines',
  SET_hideErrorOutputLines = 'globalUiReducer_SET_hideErrorOutputLines',

  RESET = 'globalUiReducer_RESET',
}


//--- console view options

export interface SET_useDiffIfOutputMismatchedAction extends ActionBase {
  readonly type: ActionType.SET_useDiffIfOutputMismatched
  readonly useDiffIfOutputMismatched: boolean
}

export interface SET_useReplaceWhiteSpacesWithBulletsAction extends ActionBase {
  readonly type: ActionType.SET_useReplaceWhiteSpacesWithBullets
  readonly useReplaceWhiteSpacesWithBullets: boolean
}

export interface SET_hideExpectedOutputLinesAction extends ActionBase {
  readonly type: ActionType.SET_hideExpectedOutputLines
  readonly hideExpectedOutputLines: boolean
}

export interface SET_hideErrorOutputLinesAction extends ActionBase {
  readonly type: ActionType.SET_hideErrorOutputLines
  readonly hideErrorOutputLines: boolean
}

//--- END console view options

export interface SET_taskDescriptionFontSizeInPxAction extends ActionBase {
  readonly type: ActionType.SET_taskDescriptionFontSizeInPx
  readonly taskDescriptionFontSizeInPx: number
}

export interface SET_quickCodeEditorFontSizeInPxAction extends ActionBase {
  readonly type: ActionType.SET_quickCodeEditorFontSizeInPx
  readonly quickCodeEditorFontSizeInPx: number | null
}


export interface ResetAction extends ActionBase {
  readonly type: ActionType.RESET
}

export type AllActions =
  ResetAction
  | SET_taskDescriptionFontSizeInPxAction
  | SET_quickCodeEditorFontSizeInPxAction
  | SET_useDiffIfOutputMismatchedAction
  | SET_useReplaceWhiteSpacesWithBulletsAction
  | SET_hideExpectedOutputLinesAction
  | SET_hideErrorOutputLinesAction

  | RESET_GlobalAction


export function reducer(state: State = initial, action: AllActions): State {

  switch (action.type) {

    case ActionType.SET_taskDescriptionFontSizeInPx:
      return {
        ...state,
        taskDescriptionFontSizeInPx: action.taskDescriptionFontSizeInPx
      }

    case ActionType.SET_quickCodeEditorFontSizeInPx:
      return {
        ...state,
        quickCodeEditorFontSizeInPx: action.quickCodeEditorFontSizeInPx
      }

    case ActionType.SET_useDiffIfOutputMismatched:
      FrontendSettingsManager.setUseDiffIfOutputMismatched(action.useDiffIfOutputMismatched)
      return {
        ...state,
        useDiffIfOutputMismatched: action.useDiffIfOutputMismatched
      }

    case ActionType.SET_useReplaceWhiteSpacesWithBullets:
      FrontendSettingsManager.setUseReplaceWhiteSpacesWithBullets(action.useReplaceWhiteSpacesWithBullets)
      return {
        ...state,
        useReplaceWhiteSpacesWithBullets: action.useReplaceWhiteSpacesWithBullets
      }

    case ActionType.SET_hideExpectedOutputLines:
      //do not store this in frontend settings because in inexperienced users will forget this to uncheck
      return {
        ...state,
        hideExpectedOutputLines: action.hideExpectedOutputLines
      }


    case ActionType.SET_hideErrorOutputLines:
      //do not store this in frontend settings because in inexperienced users will forget this to uncheck
      return {
        ...state,
        hideErrorOutputLines: action.hideErrorOutputLines
      }

    case ActionType.RESET:
    case ActionTypeGlobal.RESET_Global:
      return {
        ...initial
      }

    default:
      notExhaustive(action)
      return state
  }
}

