import {LangFullBase} from '../../../types/languages'
import {ActionBase, ActionType} from './langsViewActionTypes'
import {notExhaustive} from '../_notExhausiveHelper'
import {getNextFrontendId} from '../../../helpers/frontendIdHelper'
import Logger from '../../../helpers/logger'


import {AllActions as AllGetLangsActions, reducer as getLangsReducer} from './crud/getLangsReducer'
import {AllActions as AllCreateLangActions, reducer as createLangReducer} from './crud/createLangReducer'
import {AllActions as AllUpdateLangActions, reducer as updateLangReducer} from './crud/updateLangReducer'
import {AllActions as AllDeleteLangActions, reducer as deleteLangReducer} from './crud/deleteLangReducer'
import {ActionTypeGlobal, RESET_GlobalAction} from '../globalActions'

export type State = {

  readonly isLoading: boolean

  readonly langs: ReadonlyArray<LangFullBase>

  /**
   * the lang id we currently edit
   * null to reset
   * <= 0 to create new
   */
  readonly editingLangId: number | null

  /**
   * a copy used for editing
   * we can then rollback to original array or commit to server
   */
  readonly langCopy: LangFullBase | null
}


export const initial: State = {
  isLoading: false,
  langs: [],

  editingLangId: null,
  langCopy: null
}


export interface SET_editingLangIdAction extends ActionBase {
  readonly type: ActionType.SET_editingLangId
  readonly editingLangId: number | null
}


//--- change lang props


export interface SET_Lang_displayNameAction extends ActionBase {
  readonly type: ActionType.SET_Lang_displayName
  readonly displayName: string
}

export interface SET_Lang_langShortcutAction extends ActionBase {
  readonly type: ActionType.SET_Lang_langShortcut
  readonly langShortcut: string
}

export type AllActions =

  | AllGetLangsActions
  | AllCreateLangActions
  | AllUpdateLangActions
  | AllDeleteLangActions

  | SET_editingLangIdAction

  | SET_Lang_displayNameAction
  | SET_Lang_langShortcutAction

  | RESET_GlobalAction


export function reducer(state: State = initial, action: AllActions): State {

  switch (action.type) {

    case ActionType.GET_languages_PENDING:
    case ActionType.GET_languages_FULFILLED:
    case ActionType.GET_languages_REJECTED:
      return getLangsReducer(state, action)

    case ActionType.CREATE_lang_PENDING:
    case ActionType.CREATE_lang_FULFILLED:
    case ActionType.CREATE_lang_REJECTED:
      return createLangReducer(state, action)

    case ActionType.UPDATE_lang_PENDING:
    case ActionType.UPDATE_lang_FULFILLED:
    case ActionType.UPDATE_lang_REJECTED:
      return updateLangReducer(state, action)

    case ActionType.DELETE_lang_PENDING:
    case ActionType.DELETE_lang_FULFILLED:
    case ActionType.DELETE_lang_REJECTED:
      return deleteLangReducer(state, action)


    case ActionType.SET_editingLangId:

      //reset
      if (action.editingLangId === null) {
        return {
          ...state,
          editingLangId: null,
          langCopy: null
        }
      }

      //action.editingLangId is !== null here
      if (action.editingLangId <= 0) {
        const newId = getNextFrontendId()
        return {
          ...state,
          editingLangId: newId,
          langCopy: {
            id: newId,
            displayName: '',
            langShortcut: ''
          }
        }
      }


      //make sure we update the langs first before ending the edit

      const targetPLang = state.langs.find(p => p.id === action.editingLangId)

      if (!targetPLang) {
        Logger.warn(`could not find lang with id ${action.editingLangId} to edit`)
        return state
      }

      return {
        ...state,
        editingLangId: action.editingLangId,
        langCopy: {
          ...targetPLang
        }
      }

    case ActionType.SET_Lang_displayName:
      return {
        ...state,
        langCopy: {
          ...state.langCopy,
          displayName: action.displayName
        }
      }

    case ActionType.SET_Lang_langShortcut:
      return {
        ...state,
        langCopy: {
          ...state.langCopy,
          langShortcut: action.langShortcut
        }
      }

    case ActionTypeGlobal.RESET_Global:
      return {
        ...initial
      }

    default:
      notExhaustive(action)
      return state
  }
}

