/**
 * Created by janis dähne.
 */
import {Action} from "redux";
import * as cryptoHelper from '../../helpers/cryptoHelper'
import {notExhaustive} from "./_notExhausiveHelper";
import {ActionTypeGlobal, RESET_GlobalAction} from './globalActions'
import {FrontendSettingsManager} from '../../helpers/frontendSettingsManager'

export type State = {
  readonly username: string
  readonly password: string

  /**
   * true: is loggin in (request started)
   * false: not
   */
  readonly isLoggingIn: boolean

  /**
   * true: request started to check if the user is already logged in
   * false: not
   */
  readonly isCheckingAlreadyLoggedIn: boolean


  //--- normally this would be an extra reducer (register reducer) but too lazy
  // also we would need an additional reducer to toggle isRegisterFormDisplayed or not

  readonly isRegisterFormDisplayed: boolean

  /**
   * true: register request is called,
   * false: not
   */
  readonly isRegistering: boolean

  readonly registerToken: string
  readonly registerPw: string
  readonly registerPwAgain: string
  readonly registerFirstName: string
  readonly registerLastName: string
  readonly registerEmail: string

  /**
   * where normal = internal
   */
  readonly loginProvider: 'normal' | 'ldap'

  /**
   * an error message e.g. if the server is offline
   */
  readonly beforeLoginErrorMessage: string

}

const initial: State = {
  username: '',
  password: '', //we need this in order to clear the password input

  isLoggingIn: false,
  isCheckingAlreadyLoggedIn: false,


  isRegisterFormDisplayed: false,
  isRegistering: false,
  registerToken: '',
  registerPw: '',
  registerPwAgain: '',
  registerFirstName: '',
  registerLastName: '',
  registerEmail: '',
  beforeLoginErrorMessage: '',

  loginProvider: 'normal',
}

export enum ActionType {
  SET_USERNAME = 'LoginReducer_SET_USERNAME',
  SET_PASSWORD = 'LoginReducer_SET_PASSWORD',
  SET_isLoggingIn = 'LoginReducer_SET_isLoggingIn',
  SET_isCheckingAlreadyLoggedIn = 'LoginReducer_SET_isCheckingAlreadyLoggedIn',


  SET_isRegisterFormDisplayed = 'LoginReducer_SET_isRegisterFormDisplayed',
  SET_isRegistering = 'LoginReducer_SET_isRegistering',
  SET_registerToken = 'LoginReducer_SET_registerToken',
  SET_registerPw = 'LoginReducer_SET_registerPw',
  SET_registerPwAgain = 'LoginReducer_SET_registerPwAgain',
  SET_registerFirstName = 'LoginReducer_SET_registerFirstName',
  SET_registerLastName = 'LoginReducer_SET_registerLastName',
  SET_registerEmail = 'LoginReducer_SET_registerEmail',
  SET_beforeLoginErrorMessage = 'LoginReducer_SET_beforeLoginErrorMessage',

  SET_loginProvider = 'LoginReducer_SET_loginProvider',

  RESET = 'LoginReducer_RESET', //we need to prefix all because reset would trigger all reset actions in other reducers
}

export interface ActionBase extends Action {
  readonly type: ActionType
}

//--- register actions


export interface SET_isRegisterFormDisplayedAction extends ActionBase {
  readonly type: ActionType.SET_isRegisterFormDisplayed
  readonly isRegisterFormDisplayed: boolean
}

export interface SET_isRegisteringAction extends ActionBase {
  readonly type: ActionType.SET_isRegistering
  readonly isRegistering: boolean
}

export interface SET_registerTokenAction extends ActionBase {
  readonly type: ActionType.SET_registerToken
  readonly registerToken: string
}

export interface SET_registerPwAction extends ActionBase {
  readonly type: ActionType.SET_registerPw
  readonly registerPw: string
}

export interface SET_registerPwAgainAction extends ActionBase {
  readonly type: ActionType.SET_registerPwAgain
  readonly registerPwAgain: string
}

export interface SET_registerFirstNameAction extends ActionBase {
  readonly type: ActionType.SET_registerFirstName
  readonly registerFirstName: string
}

export interface SET_registerLastNameAction extends ActionBase {
  readonly type: ActionType.SET_registerLastName
  readonly registerLastName: string
}

export interface SET_registerEmailAction extends ActionBase {
  readonly type: ActionType.SET_registerEmail
  readonly registerEmail: string
}
export interface SET_loginProviderAction extends ActionBase {
  readonly type: ActionType.SET_loginProvider
  readonly loginProvider: 'normal' | 'ldap'
}


//--- login actions

export interface SET_isLoggingInAction extends ActionBase {
  readonly type: ActionType.SET_isLoggingIn
  readonly isLoggingIn: boolean
}

export interface SET_isCheckingAlreadyLoggedInAction extends ActionBase {
  readonly type: ActionType.SET_isCheckingAlreadyLoggedIn
  readonly isCheckingAlreadyLoggedIn: boolean
}

export interface LoginActionSetId extends ActionBase {
  readonly type: ActionType.SET_USERNAME
  readonly id: string
}

export interface LoginActionSetPw extends ActionBase {
  readonly type: ActionType.SET_PASSWORD
  readonly pw: string
}

export interface SET_beforeLoginErrorMessageAction extends ActionBase {
  readonly type: ActionType.SET_beforeLoginErrorMessage
  readonly errorMessage: string
}


export interface LoginActionReset extends ActionBase {
  readonly type: ActionType.RESET
}

export type AllActions =
  LoginActionReset
  | LoginActionSetId
  | LoginActionSetPw
  | SET_isLoggingInAction
  | SET_isCheckingAlreadyLoggedInAction

  | SET_isRegisterFormDisplayedAction
  | SET_isRegisteringAction
  | SET_registerTokenAction
  | SET_registerPwAction
  | SET_registerPwAgainAction
  | SET_registerFirstNameAction
  | SET_registerLastNameAction
  | SET_registerEmailAction
  | SET_loginProviderAction
  | SET_beforeLoginErrorMessageAction

  | RESET_GlobalAction


export function reducer(state: State = initial, action: AllActions): State {

  switch (action.type) {


    case ActionType.SET_isRegisterFormDisplayed:
      return {
        ...state,
        isRegisterFormDisplayed: action.isRegisterFormDisplayed
      }
    case ActionType.SET_isRegistering:
      return {
        ...state,
        isRegistering: action.isRegistering
      }

    case ActionType.SET_registerToken:
      return {
        ...state,
        registerToken: action.registerToken
      }

    case ActionType.SET_registerPw:
      return {
        ...state,
        registerPw: action.registerPw
      }
    case ActionType.SET_registerPwAgain:
      return {
        ...state,
        registerPwAgain: action.registerPwAgain
      }

    case ActionType.SET_registerFirstName:
      return {
        ...state,
        registerFirstName: action.registerFirstName
      }
    case ActionType.SET_registerLastName:
      return {
        ...state,
        registerLastName: action.registerLastName
      }
    case ActionType.SET_registerEmail:
      return {
        ...state,
        registerEmail: action.registerEmail
      }

    case ActionType.SET_loginProvider:
      return {
        ...state,
        loginProvider: action.loginProvider
      }


    case ActionType.SET_USERNAME:
      return {
        ...state,
        username: action.id
      }
    case ActionType.SET_PASSWORD:
      return {
        ...state,
        password: action.pw,
      }

    case ActionType.SET_isLoggingIn:
      return {
        ...state,
        isLoggingIn: action.isLoggingIn
      }

    case ActionType.SET_isCheckingAlreadyLoggedIn:
      return {
        ...state,
        isCheckingAlreadyLoggedIn: action.isCheckingAlreadyLoggedIn
      }
    case ActionType.SET_beforeLoginErrorMessage:

      //do nothing because this is set on every request...
      if (state.beforeLoginErrorMessage === action.errorMessage && action.errorMessage === '') return state

      return {
        ...state,
        beforeLoginErrorMessage: action.errorMessage
      }


    case ActionType.RESET:
    case ActionTypeGlobal.RESET_Global:
      return {
        ...initial,
        loginProvider: FrontendSettingsManager.getFrontendSettings().lastLoginProvider
      }

    default:
      notExhaustive(action)
      return state
  }
}

