/**
 * Created by janis dähne.
 */
import {notExhaustive} from "../_notExhausiveHelper";
import {NewUserFrontendExclusive, SelectableUserFrontendExclusive, UserPreviewFromBackend} from "../../../types/users";
import {ActionBase, ActionType} from "./manageNewUsersSiteActionTypes";
import {AllActions as GetActions, reducer as getReducer} from './getNewUsersReducers'
import {AllActions as DeleteActions, reducer as deleteReducer} from './deleteNewUserReducer'
import {AllActions as ActivateActions, reducer as activateReducer} from './activateNewUserReducer'
import {AllActions as GetSystemRolesActions, reducer as systemRolesReducer} from './getSystemRolesReducer'
import {SystemRoleFullBase} from "../../../types/userData";
import {SortDirection} from "../../../types/sortDirection";
import {getInitialPaginationData} from '../../../constants'
import {PaginationStateExtend} from '../../../types/pagination'
import {FrontendSettingsManager} from '../../../helpers/frontendSettingsManager'
import {ActionTypeGlobal, RESET_GlobalAction} from '../globalActions'


export type State = {
  /**
   * the activated users
   */
  readonly newUsers: ReadonlyArray<NewUserFrontendExclusive>
  readonly systemRoles: ReadonlyArray<SystemRoleFullBase>
  readonly isLoading: boolean

} & PaginationStateExtend<UserPreviewFromBackend>

export const initial: State = {
  newUsers: [],
  systemRoles: [],
  isLoading: false,

  pagination: {
    ...getInitialPaginationData(),
    orderByProp: 'lastName'
  }
}

export interface SetSearchTextAction extends ActionBase {
  readonly type: ActionType.setSearchText
  readonly searchText: string
}

export interface SetActivePaginationPageAction extends ActionBase {
  readonly type: ActionType.setPaginationActivePage
  readonly activePage: number
}

export interface SetPaginationPageSizeAction extends ActionBase {
  readonly type: ActionType.setPaginationPageSize
  readonly pageSize: number
}

export interface SET_sortByKeyAction extends ActionBase {
  readonly type: ActionType.SET_sortByKey
  readonly sortByKey: keyof UserPreviewFromBackend
  readonly sortDirection: boolean | null
}

export interface SET_newUsersAction extends ActionBase {
  readonly type: ActionType.SET_newUsers
  readonly newUsers: ReadonlyArray<NewUserFrontendExclusive>
}

export interface SET_newUser_systemRoleAction extends ActionBase {
  readonly type: ActionType.SET_newUser_systemRole
  readonly newUser: NewUserFrontendExclusive
  readonly systemRole: SystemRoleFullBase
}

export interface SET_newUser_uiIsChangeSystemRolePopupDisplayedAction extends ActionBase {
  readonly type: ActionType.SET_newUser_uiIsChangeSystemRolePopupDisplayed
  readonly newUserId: number
  readonly isDisplayed: boolean
}

export interface SET_newUser_uiIsSelectedAction extends ActionBase {
  readonly type: ActionType.SET_newUser_uiIsSelected
  readonly newUserId: number
  readonly isSelected: boolean
}

export interface SET_all_newUsers_uiIsSelectedAction extends ActionBase {
  readonly type: ActionType.SET_all_newUsers_uiIsSelected
  readonly isSelected: boolean
}


export interface SET_isLoadingAction extends ActionBase {
  readonly type: ActionType.SET_isLoading
  readonly isLoading: boolean
}

export interface ResetAction extends ActionBase {
  readonly type: ActionType.RESET
}

export type AllActions =
  ResetAction
  | SET_newUsersAction
  | SET_newUser_systemRoleAction
  | SET_isLoadingAction
  | GetActions
  | DeleteActions
  | ActivateActions
  | GetSystemRolesActions
  | SET_newUser_uiIsChangeSystemRolePopupDisplayedAction
  | SET_newUser_uiIsSelectedAction
  | SET_all_newUsers_uiIsSelectedAction
  | SET_sortByKeyAction
  | SetSearchTextAction
  | SetActivePaginationPageAction
  | SetPaginationPageSizeAction

  | RESET_GlobalAction

export function reducer(state: State = initial, action: AllActions): State {

  switch (action.type) {

    case ActionType.SET_sortByKey:
      return {
        ...state,
        pagination: {
          ...state.pagination,
          orderByProp: action.sortByKey,
          isOrderByDesc: action.sortDirection
        }
      }

    case ActionType.GET_newUsers_PENDING:
    case ActionType.GET_newUsers_FULFILLED:
    case ActionType.GET_newUsers_REJECTED:
      return getReducer(state, action)

    case ActionType.DELETE_newUser_PENDING:
    case ActionType.DELETE_newUser_FULFILLED:
    case ActionType.DELETE_newUser_REJECTED:
      return deleteReducer(state, action)

    case ActionType.ACTIVATE_newUser_PENDING:
    case ActionType.ACTIVATE_newUser_FULFILLED:
    case ActionType.ACTIVATE_newUser_REJECTED:
      return activateReducer(state, action)

    case ActionType.GET_systemRoles_PENDING:
    case ActionType.GET_systemRoles_FULFILLED:
    case ActionType.GET_systemRoles_REJECTED:
      return systemRolesReducer(state, action)

    case ActionType.setSearchText:
      return {
        ...state,
        pagination: {
          ...state.pagination,
          searchText: action.searchText
        }
      }
    case ActionType.setPaginationActivePage:
      return {
        ...state,
        pagination: {
          ...state.pagination,
          currentPage: action.activePage
        }
      }

    case ActionType.setPaginationPageSize:

      FrontendSettingsManager.set_manageNewUsersPageSize(action.pageSize)

      return {
        ...state,
        pagination: {
          ...state.pagination,
          pageSize: action.pageSize,
          currentPage: 1,
        }
      }

    case ActionType.SET_newUsers:
      return {
        ...state,
        newUsers: action.newUsers
      }

    case ActionType.SET_newUser_systemRole:
      return {
        ...state,
        newUsers: state.newUsers.map(
          user => user.id === action.newUser.id
                  ?
            {
              ...action.newUser,
              systemRoleId: (action.systemRole === undefined
                             ? null
                             : action.systemRole.id),
              systemRole: action.systemRole
            }
                  : user
        )
      }

    case ActionType.SET_newUser_uiIsSelected:
      return {
        ...state,
        newUsers: state.newUsers.map(
          user => user.id === action.newUserId
                  ?
            {
              ...user,
              uiIsSelected: action.isSelected
            }
                  : user)
      }

    case ActionType.SET_all_newUsers_uiIsSelected:
      return {
        ...state,
        newUsers: state.newUsers.map(user => {
          return {
            ...user,
            uiIsSelected: action.isSelected
          }
        })
      }

    case ActionType.SET_newUser_uiIsChangeSystemRolePopupDisplayed:
      return {
        ...state,
        newUsers: state.newUsers.map(
          user => user.id === action.newUserId
                  ?
            {
              ...user,
              uiIsChangeSystemRolePopupDisplayed: action.isDisplayed
            }
                  : user)
      }

    case ActionType.SET_isLoading:
      return {
        ...state,
        isLoading: action.isLoading
      }

    case ActionType.RESET:
    case ActionTypeGlobal.RESET_Global:
      return {
        ...initial,
        pagination: {
          ...initial.pagination,
          pageSize: FrontendSettingsManager.getFrontendSettings().manageNewUsersPageSize
        }
      }

    default:
      notExhaustive(action)
      return state
  }
}

