/**
 * Created by janis dähne.
 */
import {EditableExercisePreviewFromBackend} from "../../../types/exercisePreview";
import {notExhaustive} from "../_notExhausiveHelper";
import {ActionBase, ActionType} from "./ownExercisesSiteActionTypes";
import {AllActions as GetOwnExercisesActions, reducer as getOwnExercisesReducer} from './getOwnExercisesReducer'
import {SortDirection} from "../../../types/sortDirection";
import {AllActions as DeleteOwnExerciseActions, reducer as deleteOwnExerciseReducer} from './deleteOwnExerciseReducer'
import {FrontendSettingsManager} from "../../../helpers/frontendSettingsManager";
import {PaginationStateExtend} from '../../../types/pagination'
import {getInitialPaginationData} from '../../../constants'
import {ActionTypeGlobal, RESET_GlobalAction} from '../globalActions'

export type State = {
  readonly ownExercises: ReadonlyArray<EditableExercisePreviewFromBackend>
  readonly isLoading: boolean

  readonly sortByKey: keyof EditableExercisePreviewFromBackend
  readonly sortDirection: SortDirection

  readonly isTagsFilterDisplayed: boolean

} & PaginationStateExtend<EditableExercisePreviewFromBackend>


export const initial: State = {
  ownExercises: [],
  isLoading: false,

  sortByKey: 'lastUpdatedAt',
  sortDirection: SortDirection.descending,

  isTagsFilterDisplayed: false,

  pagination: {
    ...getInitialPaginationData(),
    orderByProp: 'lastUpdatedAt',
    isOrderByDesc: true,
  }
}


export interface SetIsLoadingAction extends ActionBase {
  readonly type: ActionType.setIsLoading
  readonly isLoading: boolean
}

//---pagination
export interface SetSearchTextAction extends ActionBase {
  readonly type: ActionType.setSearchText
  readonly searchText: string
}

export interface SetActivePaginationPageAction extends ActionBase {
  readonly type: ActionType.setPaginationActivePage
  readonly activePage: number
}

export interface SetPaginationPageSizeAction extends ActionBase {
  readonly type: ActionType.setPaginationPageSize
  readonly pageSize: number
}

//---pagination end

export interface SET_sortByKeyAction extends ActionBase {
  readonly type: ActionType.SET_sortByKey
  readonly sortByKey: keyof EditableExercisePreviewFromBackend
  readonly sortDirection: boolean | null
}

export interface SET_isTagsFilterDisplayedAction extends ActionBase {
  readonly type: ActionType.SET_isTagsFilterDisplayed
  readonly isTagsFilterDisplayed: boolean
}

export interface ResetAction extends ActionBase {
  readonly type: ActionType.RESET
  /**
   * really annoying if you go back and have to search/sort again...
   */
  readonly resetPaginationState: boolean
}

export type AllActions =
  ResetAction
  | SetIsLoadingAction
  | SET_sortByKeyAction
  | SET_isTagsFilterDisplayedAction

  | DeleteOwnExerciseActions
  | GetOwnExercisesActions
  | SetSearchTextAction
  | SetActivePaginationPageAction
  | SetPaginationPageSizeAction

  | RESET_GlobalAction

export function reducer(state: State = initial, action: AllActions): State {

  switch (action.type) {

    case ActionType.setIsLoading:
      return {
        ...state,
        isLoading: action.isLoading
      }

    case ActionType.GET_ownExercises_PENDING:
    case ActionType.GET_ownExercises_FULFILLED:
    case ActionType.GET_ownExercises_REJECTED:
      return getOwnExercisesReducer(state, action)

    case ActionType.DELETE_ownExercise_PENDING:
    case ActionType.DELETE_ownExercise_FULFILLED:
    case ActionType.DELETE_ownExercise_REJECTED:
      return deleteOwnExerciseReducer(state, action)

    case ActionType.SET_isTagsFilterDisplayed:

      FrontendSettingsManager.setOwnExercisesSiteTagsFilterIsDisplayed(action.isTagsFilterDisplayed)

      return {
        ...state,
        isTagsFilterDisplayed: action.isTagsFilterDisplayed
      }

    case ActionType.SET_sortByKey:
      return {
        ...state,
        pagination: {
          ...state.pagination,
          orderByProp: action.sortByKey,
          isOrderByDesc: action.sortDirection
        }
      }

    case ActionType.setSearchText:
      return {
        ...state,
        pagination: {
          ...state.pagination,
          searchText: action.searchText
        }
      }
    case ActionType.setPaginationActivePage:
      return {
        ...state,
        pagination: {
          ...state.pagination,
          currentPage: action.activePage
        }
      }

    case ActionType.setPaginationPageSize:

      FrontendSettingsManager.set_ownExercisesPageSize(action.pageSize)

      return {
        ...state,
        pagination: {
          ...state.pagination,
          pageSize: action.pageSize,
          currentPage: 1,
        }
      }

    case ActionTypeGlobal.RESET_Global: {
      return {
        ...initial,
        isTagsFilterDisplayed: FrontendSettingsManager.getFrontendSettings().ownExercisesSiteTagsFilterIsDisplayed,
        pagination: {
          ...initial.pagination,
          pageSize: FrontendSettingsManager.getFrontendSettings().ownExercisesPageSize
        }
      }
    }
    case ActionType.RESET: {

      if (!action.resetPaginationState) {
        return {
          ...initial,
          sortByKey: state.sortByKey,
          sortDirection: state.sortDirection,

          pagination: {
            ...state.pagination
          }
        }
      }

      return {
        ...initial,
        isTagsFilterDisplayed: FrontendSettingsManager.getFrontendSettings().ownExercisesSiteTagsFilterIsDisplayed,
        pagination: {
          ...initial.pagination,
          pageSize: FrontendSettingsManager.getFrontendSettings().ownExercisesPageSize
        }
      }
    }


    default:
      notExhaustive(action)
      return state
  }
}

