/**
 * Created by janis dähne.
 */

import {notExhaustive} from "../../_notExhausiveHelper";
import {EditableExercisePreviewFromBackend} from "../../../../types/exercisePreview";
import {AllActions as GetReleasesAction, reducer as getReleasesReducer} from './getReleasesReducer'
import {ActionBase, ActionType} from "./releasesSiteActionTypes";
import {ExerciseReleaseFromBackend} from "../../../../types/exerciseRelease";
import {
  AllActions as GetOwnExercisesActions,
  reducer as getSingleOwnExerciseCrudReducer
} from './getSingleOwnExerciseCrudReducer'
import {
  AllActions as GetGroupExercisesActions,
  reducer as getSingleGroupExerciseCrudReducer
} from './getSingleGroupExerciseCrudReducer'
import {AllActions as DeleteExerciseReleaseActions, reducer as deleteReleaseReducer} from './deleteReleaseReducer'
import {SortDirection} from "../../../../types/sortDirection";
import {GroupRoleWithPermissionFullBase} from "../../../../types/groupRoles";
import {PaginationStateExtend} from '../../../../types/pagination'
import {getInitialPaginationData, paginationPageSizeOptionAll} from '../../../../constants'
import {GroupPreviewFromBackend} from '../../../../types/group'
import {FrontendSettingsManager} from '../../../../helpers/frontendSettingsManager'
import {ActionTypeGlobal, RESET_GlobalAction} from '../../globalActions'


export type State = {

  /**
   * true: we use the own exercises (maybe we need to load them), we don't need the group role, we have the permission
   * false: we use the group exercises (maybe we need to load them)
   */
  readonly comesFromOwnExercisesSite: boolean

  /**
   * is null if we comesFromOwnExercisesSite  but also on error
   */
  readonly group: GroupPreviewFromBackend | null

  /**
   * is null if we comesFromOwnExercisesSite but also on error
   */
  readonly groupRole: GroupRoleWithPermissionFullBase | null

  readonly selectedExerciseId: number

  /**
   * null on error or init
   */
  readonly exercise: EditableExercisePreviewFromBackend | null
  readonly releases: ReadonlyArray<ExerciseReleaseFromBackend>

  readonly isLoading: boolean


  readonly isZoomedExerciseReleaseDisplayed: boolean

  readonly isChangeDialogDisplayed: boolean

  readonly sortByKey: keyof ExerciseReleaseFromBackend
  readonly sortDirection: SortDirection

} & PaginationStateExtend<ExerciseReleaseFromBackend>

export const initial: State = {
  exercise: null,
  group: null,
  groupRole: null,
  selectedExerciseId: -1,
  releases: [],
  isLoading: false,
  comesFromOwnExercisesSite: false,
  isZoomedExerciseReleaseDisplayed: false,
  isChangeDialogDisplayed: false,

  sortByKey: 'id',
  sortDirection: SortDirection.ascending,

  pagination: {
    ...getInitialPaginationData(),
    orderByProp: 'releasedAt',
    isOrderByDesc: true,
  }
}


//---pagination
export interface SetSearchTextAction extends ActionBase {
  readonly type: ActionType.setSearchText
  readonly searchText: string
}

export interface SetActivePaginationPageAction extends ActionBase {
  readonly type: ActionType.setPaginationActivePage
  readonly activePage: number
}

export interface SetPaginationPageSizeAction extends ActionBase {
  readonly type: ActionType.setPaginationPageSize
  readonly pageSize: number
}

//---pagination end

export interface SET_sortByKeyAction extends ActionBase {
  readonly type: ActionType.SET_sortByKey
  readonly sortByKey: keyof ExerciseReleaseFromBackend
  readonly sortDirection: boolean | null
}

export interface SET_comesFromOwnExercisesSiteAction extends ActionBase {
  readonly type: ActionType.SET_comesFromOwnExercisesSite
  readonly comesFromOwnExercisesSite: boolean
}

export interface SET_selectedExerciseIdAction extends ActionBase {
  readonly type: ActionType.SET_selectedExerciseId
  readonly id: number
}

export interface SET_isZoomedExerciseReleaseDisplayedAction extends ActionBase {
  readonly type: ActionType.SET_isZoomedExerciseReleaseDisplayed
  readonly isDisplayed: boolean
}

export interface SET_isChangeDialogDisplayedAction extends ActionBase {
  readonly type: ActionType.SET_isChangeDialogDisplayed
  readonly isDisplayed: boolean
}


export interface ResetAction extends ActionBase {
  readonly type: ActionType.RESET
  /**
   * really annoying if you go back and have to search/sort again...
   */
  readonly resetPaginationState: boolean
}

export type AllActions =
  ResetAction
  | SET_selectedExerciseIdAction

  | SET_comesFromOwnExercisesSiteAction
  | SET_isZoomedExerciseReleaseDisplayedAction
  | SET_isChangeDialogDisplayedAction
  | SET_sortByKeyAction

  | DeleteExerciseReleaseActions
  | GetReleasesAction
  | GetOwnExercisesActions
  | GetGroupExercisesActions
  | SetSearchTextAction
  | SetActivePaginationPageAction
  | SetPaginationPageSizeAction

  | RESET_GlobalAction

export function reducer(state: State = initial, action: AllActions): State {

  switch (action.type) {


    case ActionType.DELETE_exerciseRelease_PENDING:
    case ActionType.DELETE_exerciseRelease_FULFILLED:
    case ActionType.DELETE_exerciseRelease_REJECTED:
      return deleteReleaseReducer(state, action)

    case ActionType.GET_exerciseReleases_PENDING:
    case ActionType.GET_exerciseReleases_FULFILLED:
    case ActionType.GET_exerciseReleases_REJECTED:
      return getReleasesReducer(state, action)


    case ActionType.GET_singleOwnExercise_PENDING:
    case ActionType.GET_singleOwnExercise_FULFILLED:
    case ActionType.GET_singleOwnExercise_REJECTED:
      return getSingleOwnExerciseCrudReducer(state, action)


    case ActionType.GET_singleGroupExercise_PENDING:
    case ActionType.GET_singleGroupExercise_FULFILLED:
    case ActionType.GET_singleGroupExercise_REJECTED:
      return getSingleGroupExerciseCrudReducer(state, action)


    case ActionType.SET_isZoomedExerciseReleaseDisplayed:
      return {
        ...state,
        isZoomedExerciseReleaseDisplayed: action.isDisplayed,
        isChangeDialogDisplayed: false //just to be sure
      }
    case ActionType.SET_isChangeDialogDisplayed:
      return {
        ...state,
        isZoomedExerciseReleaseDisplayed: false,
        isChangeDialogDisplayed: action.isDisplayed
      }


    case ActionType.SET_comesFromOwnExercisesSite:
      return {
        ...state,
        comesFromOwnExercisesSite: action.comesFromOwnExercisesSite
      }

    case ActionType.SET_selectedExerciseId:
      return {
        ...state,
        selectedExerciseId: action.id
      }

    case ActionType.SET_sortByKey:
      return {
        ...state,
        pagination: {
          ...state.pagination,
          orderByProp: action.sortByKey,
          isOrderByDesc: action.sortDirection
        }
      }

    case ActionType.setSearchText:
      return {
        ...state,
        pagination: {
          ...state.pagination,
          searchText: action.searchText
        }
      }
    case ActionType.setPaginationActivePage:
      return {
        ...state,
        pagination: {
          ...state.pagination,
          currentPage: action.activePage
        }
      }

    case ActionType.setPaginationPageSize:

      FrontendSettingsManager.set_releasesViewPageSize(action.pageSize)

      return {
        ...state,
        pagination: {
          ...state.pagination,
          pageSize: action.pageSize,
          currentPage: 1,
        }
      }

    case ActionTypeGlobal.RESET_Global: {
      return {
        ...initial,
        pagination: {
          ...initial.pagination,
          pageSize: FrontendSettingsManager.getFrontendSettings().releasesViewPageSize
        }
      }
    }
    case ActionType.RESET: {

      if (!action.resetPaginationState) {
        return {
          ...initial,
          sortByKey: state.sortByKey,
          sortDirection: state.sortDirection,

          //...this might be bad if the server side changed (e.g. deleted/added exercises)
          pagination: {
            ...state.pagination
          }
        }
      }

      return {
        ...initial,
        pagination: {
          ...initial.pagination,
          pageSize: FrontendSettingsManager.getFrontendSettings().releasesViewPageSize
        }
      }
    }



    default:
      notExhaustive(action)
      return state
  }
}

