/**
 * Created by janis dähne.
 */

import {availableEditorThemeNames, InternalCodeEditorThemes} from "../helpers/themes";

export type ValidationFunc<T> = (value: T) => boolean


//notice the power of typescript!!, see https://www.typescriptlang.org/docs/handbook/release-notes/typescript-2-1.html

//TODO
//this is nice for after convert validation
//but e.g. for numbers this is bad because user inputs a string not a number
//using the state property type is good for enums (checkbox, radio button stuff) because there is no wrong input
export type ValidationRulesCollection<T> = {[fieldName in keyof T]: ReadonlyArray<ValidationFunc<T[fieldName]>> }

export type ValidationMessageKeysCollection<T> = {[fieldName in keyof T]: string }


/**
 * just a helper to only put valid field names in the collection object
 *
 * usage:
 *
 * const validations = getValidationCollection<SystemRoleFullBase>({
 *   id: [() => true],
 *   displayName: [() => true],,
 *   email: false,
 *   canCreateGroups: [() => true],,
 *   canChangeGroups: [() => true],,
 *   canDeleteGroups: [() => true],
 *   ...
 * })
 *
 * compiler will warn if not all field have an entry or if an entry is not an field in the SystemRoleFullBase type
 * @param {T} collectionObject
 * @returns {T}
 */
export function getValidationCollection<T>(collectionObject: ValidationRulesCollection<T>): ValidationRulesCollection<T> {
  return collectionObject
}


/**
 * a helper to specify the validation messages
 *
 * usage:
 *
 * export const validationMessageKeys = getValidationMessagesCollection<State>({
 *   id: '',
 *   displayName: 'TODO_translate_key',
 *   email: 'TODO_invalid',
 *
 *   canCreateGroups: '',
 *   canChangeGroups: '',
 *   canDeleteGroups: '',
 *
 *   canCreateRoles: '',
 *   canChangeRoles: '',
 *   canDeleteRoles: '',
 *
 *   canManageNewUsers: '',
 *   canChangeOtherUsersSystemRole: '',
 *
 *   canDeleteActivatedUsers: '',
 *   canChangeUserData: '',
 *   canChangeSystemSettings: '',
 *   canManageTags: '',
 *   isValid: ''
 * })
 *
 * @param {ValidationMessageKeysCollection<T>} collectionObject
 * @returns {ValidationMessageKeysCollection<T>}
 */
export function getValidationMessagesCollection<T>(collectionObject: ValidationMessageKeysCollection<T>): ValidationMessageKeysCollection<T> {
  return collectionObject
}


//from https://github.com/arasatasaygin/is.js/blob/master/is.js
const emailRegex = /^((([a-z]|\d|[!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+(\.([a-z]|\d|[!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+)*)|((\x22)((((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(([\x01-\x08\x0b\x0c\x0e-\x1f\x7f]|\x21|[\x23-\x5b]|[\x5d-\x7e]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(\\([\x01-\x09\x0b\x0c\x0d-\x7f]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]))))*(((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(\x22)))@((([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))\.)+(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))$/i

export function isNotNullOrUndefined(text: any): boolean {
  return text !== null && text !== undefined
}

export function isNotSpaces(text: string): boolean {
  return text.trim() !== ''
}


/**
 * checks whether the given text is an email
 * @param {string} text
 * @returns {boolean}
 */
export function isEmail(text: string): boolean {
  return emailRegex.test(text)
}

export function isFlatEqual<T>(item1: T, item2: T): boolean {
  return item1 === item2
}

/**
 * email or empty (see isNotEmpty)
 * @param {string} text
 * @returns {boolean}
 */
export function isEmailOrEmpty(text: string): boolean {
  return !isNotEmpty(text) || emailRegex.test(text)
}

/**
 * email and not empty (see isNotEmpty)
 * @param {string} text
 * @returns {boolean}
 */
export function isEmailAndNotEmpty(text: string): boolean {
  return isNotEmpty(text) && emailRegex.test(text)
}

/**
 * checks whether the given text not empty ('' or null or undefined)
 * @param {string} text
 * @returns {boolean}
 */
export function isNotEmpty(text: string): boolean {
  return text !== null && text !== undefined && text !== ''
}

/**
 * checks whether the given text not empty or only whitespaces ('' or null or undefined)
 * @param {string} text
 * @returns {boolean}
 */
export function isNotEmptyOrSpaces(text: string): boolean {
  return text !== null && text !== undefined && text.trim() !== ''
}

export function isEmptyOrNotSpaces(text: string): boolean {
  return !isNotNullOrUndefined(text) || text.length === 0 || isNotSpaces(text)
}

export function isIdString(id: string): boolean {

  if (isNotNullOrUndefined(id)) return false
  // noinspection TsLint
  const check = parseInt(id, 10)
  return !isNaN(check) && check > 0;
}

export function isValidId(id: number): boolean {

  if (isNaN(id)) return false
  return id > 0
}

export function isNotNan(num: number): boolean {
  return !isNaN(num)
}


/**
 * checks if the given number is a real number and >= 0 (and not NaN)
 * @param {number | null | undefined} num
 * @returns {boolean}
 */
export function isPositiveNumber(num: number | null | undefined): boolean {

  if (num === null || num === undefined) return false

  if (isNaN(num)) return false

  return num > 0;


}

export function isValidEditorTheme(themeValue: string): boolean {

  for (const key in InternalCodeEditorThemes) {
    if (InternalCodeEditorThemes.hasOwnProperty(key)) {
      if (key === themeValue) return true
    }
  }

  return false
}